#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Apr 23 21:50:49 2021

This module is used to specify the metrics used for evaluation of the properties of molecules of library.

@author: tucy
"""

import numpy as np
import pandas as pd
from rdkit import Chem
from rdkit.Chem import BRICS
from salam.module_MD.OrganoMetalics_Enumeration import find_CH_aromatic_idxs

from salam.module_MD.SA_Score import sascorer
from rdkit.Chem import Draw
from rdkit.Chem import PandasTools
import matplotlib.pyplot as plt

import re

import subprocess
from salam.module_MD.D_Pi_A_Enumeration import write_mols_paralell

from rdkit import DataStructs
import matplotlib.image as mpimg # mpimg 用于读取图片


#-----------------------------------------------------------------------------

def convert_str_to_list(str1="[ '*F',  '*C#N',  '*OC',  '*N(C)C'  ]"):
    # convert a str, "[ '*F',  '*C#N',  '*OC',  '*N(C)C'  ]"  to a list, ['*F', '*C#N', '*OC', '*N(C)C']
    print( "type(str1) = ", type(str1) )
    print( str1 )
    
    str1_list =  str1.strip().replace('[', '').replace(']', '').split(sep=',')
    
    list1 =  [ x.strip().replace('"', '').replace("'", '')  for x in  str1_list ]
    
    print("len(list1) = ", len(list1))

    print("type(list1) = ", type(list1))
    print(list1)
    
    return list1


def get_paras_dic( filename='./salam_paras.inp' ):
    print("input file path is:")
    print(filename)
    print("# Begin parsing paras.\n")
    paras_dic = {}
    #comment_pattern = re.compile(r'#(.*?)', re.MULTILINE|re.DOTALL)
    comment_pattern = re.compile(r'^#(.*?)')
    blankline_pattern = re.compile(r'^(\s*?)$')
    # line_no = 1
    with open(filename, 'r', encoding='UTF-8') as fr:
        for line in fr:   
            comment_result = comment_pattern.findall(line)
            blankline_result = blankline_pattern.findall(line)
            
            if ((len(comment_result) > 0) or (len(blankline_result) > 0)):
                # print(line_no, " COMMENT LINE FOUND ", line)
                print(line, end='')
            else:
                # print(line_no, "  ", line)
                print(line, end='')
                values = line.strip().split('=')  #通过：分割字符串变成一个列表    
                # print("values: ", values)    
                new_values = [ val.strip() for val in values]
                print(new_values)   
                #paras_dic[values[0]] = values[1]
                paras_dic[new_values[0]] = new_values[1]
            
            # line_no += 1
    
    print('-'*60)
    print("paras_dic.items():")
    print(paras_dic.items() )
    
    print("\n# End of parsing paras.\n")

    return paras_dic


#GXXXX_df_ECFP_csv_path = './project/'  +  GXXXX    +  '/'  + GXXXX   +  '_df_ECFP.csv'

def percentage_lt_criticval(filename='./project/G0000/G0000_df_ECFP.csv', 
                            critical_value=0.15):
    
    df_ECFP = pd.read_csv(filename)
    print("df_ECFP.head():\n")
    print(df_ECFP.head())
    
    df_ECFP['prop_preds_is_lower_than_threshold'] = df_ECFP['prop_preds'].apply( lambda x: True if (x <= critical_value) else False )
    
    #df_ECFP_prop_low = df_ECFP.loc[ df_ECFP['prop_preds'] <= critical_value ] 
    
    #total_num = len(df_ECFP['prop_preds'])
    total_num = df_ECFP.shape[0]
    passed_num = df_ECFP['prop_preds_is_lower_than_threshold'].sum()
    
    if ( total_num >= 1):
        percentage = passed_num / float( total_num )
    else:
        print("The 1st shape of the dataframe,  df_ECFP.shape[0] must >= 1.\n")
        exit(-1)
    
    return percentage


def percentage_lt_criticval_df(df, 
                               colname='prop_preds',
                               critical_value=0.15):
    
    print("df.head():\n")
    print(df.head())
    
    df[ '%s_is_lower_than_threshold'%colname ] = df[ colname ].apply( lambda x: True if (x <= critical_value) else False )
    
    #total_num = len(df_ECFP['prop_preds'])
    total_num = df.shape[0]
    passed_num = df[ '%s_is_lower_than_threshold'%colname ].sum()
    
    if ( total_num >= 1):
        percentage = passed_num / float( total_num )
    else:
        print("The 1st shape of the dataframe,  df.shape[0] must >= 1.\n")
        exit(-1)
    
    return percentage



def getmaxminidxs_lt_le_gt_ge_criticval_df(df, 
                                           colname='prop_preds',
                                           critical_value=0.15,
                                           ascending=True):
    df_ = df.copy(deep=True)
    print("df_.head():\n")
    print(df_.head())
    
    df_[ '%s_is_lt_tsd'%colname ] = df_[ colname ].apply( lambda x: True if (x < critical_value) else False )
    df_[ '%s_is_le_tsd'%colname ] = df_[ colname ].apply( lambda x: True if (x <= critical_value) else False )
    df_[ '%s_is_gt_tsd'%colname ] = df_[ colname ].apply( lambda x: True if (x > critical_value) else False )
    df_[ '%s_is_ge_tsd'%colname ] = df_[ colname ].apply( lambda x: True if (x >= critical_value) else False )
    
    print("df_.head():\n")
    print(df_.head())
    
    lt_num = df_[ '%s_is_lt_tsd'%colname ].sum()
    le_num = df_[ '%s_is_le_tsd'%colname ].sum()
    gt_num = df_[ '%s_is_gt_tsd'%colname ].sum()
    ge_num = df_[ '%s_is_ge_tsd'%colname ].sum()
    
    
    print("lt_num = %8d, le_num = %8d, gt_num = %8d, ge_num = %8d"%(lt_num, le_num, gt_num, ge_num) )
    if (lt_num * le_num * gt_num * ge_num == 0):
        print("Warning: one of lt_num, le_num, gt_num, ge_num = 0. Use with careness.")
    
    if ( ascending==True ):
        lt_idxmax = lt_num
        le_idxmax = le_num
        gt_idxmin = le_num + 1
        ge_idxmin = lt_num + 1
        print("order is ascending. Return, lt_idxmax, le_idxmax, gt_idxmin, ge_idxmin")
        return lt_idxmax, le_idxmax, gt_idxmin, ge_idxmin
    
    else:
        lt_idxmin = ge_num + 1
        le_idxmin = gt_num + 1
        gt_idxmax = gt_num
        ge_idxmax = ge_num
        print("order is descending. Return, lt_idxmin, le_idxmin, gt_idxmax, ge_idxmax")
        return lt_idxmin, le_idxmin, gt_idxmax, ge_idxmax



def get_allfrags_via_bricsdecompose(filename='./project/G0000/G0000.sdf'):
    #----- generating fragments from dataset of mols by BRICSDecompose -----------
    mols = Chem.SDMolSupplier( filename )
    print("len(mols) = ", len(mols))
    
    allfrags = set()
    for mol in mols:
        pieces = BRICS.BRICSDecompose(mol)
        allfrags.update(pieces)
    
    print("len(allfrags) = ", len(allfrags) )
    
    return list( sorted(allfrags) )


def get_num_replaceable_aromatic_CHs(filename='./project/G0000/G0000.sdf', 
                                     patt_str="c[H]"
                                     ):
    #----- compute number of replaceable aromatic CHs from dataset of mols -----------
    mols = Chem.SDMolSupplier( filename )
    
    print("len(mols) = ", len(mols))
    
    num_replaceable_aromatic_CHs = []
    for mol in mols:
        mol = Chem.AddHs(mol)
        Xatom_label_idxs, H_Xatom_label_idxs = find_CH_aromatic_idxs(mol, 
                                                                     patt=Chem.MolFromSmarts( patt_str ) 
                                                                     )
        #print( "len(Xatom_label_idxs) = ",  len(Xatom_label_idxs) )
        num_replaceable_aromatic_CHs.append( len(Xatom_label_idxs) )
        
    
    print("sum(num_replaceable_aromatic_CHs) = ", sum(num_replaceable_aromatic_CHs) )
    print("len(num_replaceable_aromatic_CHs) = ", len(num_replaceable_aromatic_CHs) )
    
    if len(mols) == 0:
        ave_num_replaceable_aromatic_CHs = 0.
    else:
        ave_num_replaceable_aromatic_CHs = sum( num_replaceable_aromatic_CHs )/float(len(mols))
    
    print("ave_num_replaceable_aromatic_CHs = ", ave_num_replaceable_aromatic_CHs)
    
    return  ave_num_replaceable_aromatic_CHs


def compute_plot_SAScore_vs_generation(filename='./project/G0000/G0000_df_ECFP.csv',
                                        column_names=[ 'compoundnames', 'smiles', 'prop_preds' ], 
                                        figname='./foo--sascores.png',
                                        GXXXX='G0000'
                                        ):
    
    df = pd.read_csv(filename, 
                     usecols=column_names
                     #names=['new_compoundnames', 'new_mols', 'new_smiles', 'new_prop_preds']
                     )

    #df = pd.read_csv('./SA_Score/Compounds_smiles.csv', sep=',')
    
    print( "\ndf.head()" )
    print( df.head() )
    print( "\ndf.describe()" )
    print( df.describe() )
    print( "\ndf.columns" )
    print( df.columns )
    
    #将smiles转换为RDKit 的Mol对象
    
    PandasTools.AddMoleculeColumnToFrame(frame=df, smilesCol='smiles')
    
    print( "\ndf.head()" )
    print( df.head() )
    print( "\ndf.describe()" )
    print( df.describe() )
    print( "\ndf.columns" )
    print( df.columns )
    
    
    df['calc_SA_score'] = df.ROMol.map(sascorer.calculateScore)
    
    # x = df.compoundnames
    # y = df.calc_SA_score
    # #with mpl.style.context('seaborn'):
    # plt.plot(x, y, '-')
    # plt.xlabel('compoundnames')
    # plt.ylabel('calculated SA scores')
    # plt.title('SA scores for compounds')
    
    # (id_max, id_min) = (df.calc_SA_score.idxmax(), df.calc_SA_score.idxmin())
    # sa_mols = [df.ROMol[id_max], df.ROMol[id_min]]
    # Draw.MolsToGridImage(sa_mols, subImgSize=(340,200),
    #                      legends=['SA-score: {:.2f}'.format(df.calc_SA_score[i]) for i in [id_max, id_min]])
    
    print( "\ndf.head()" )
    print( df.head() )
    print( "\ndf.describe()" )
    print( df.describe() )
    print( "\ndf.columns" )
    print( df.columns )
    
    #df['calc_SA_score'].plot( ) 

    fig1, ax1 = plt.subplots(1, 1, figsize=(12, 8), sharey=False)
    
    df['calc_SA_score'].plot(kind='hist',
                             title='distribution of calc_SA_score in %s'%GXXXX,
                             #color='r', 
                             ax=ax1, 
                             ) 
    
    # df_generation_abundance_aromaticCHs.plot(
    #                                         title='aromaticCH vs generation',
    #                                         #kind='scatter', 
    #                                         kind='line', linestyle='-',
    #                                         x='generation', 
    #                                         y='aromaticCH', 
    #                                         color='r', 
    #                                         alpha=0.95, 
    #                                         ax=ax1, 
    #                                         xlabel='generation',
    #                                         ylabel='aromaticCH'
    #                                         )
    
    ax1.legend(ncol=1, fontsize='x-large', shadow=True)
    ax1.grid(True)
    #fig1.show()
    fig1.savefig(figname)

    
    return df['calc_SA_score'].tolist()


def compute_plot_prop_masscenter_vs_generation(filenames=[ './project/%s/%s_df_ECFP.csv'%(GXXXX, GXXXX) for GXXXX in ['G0000', 'G0001', 'G0002', 'G0003', 'G0004', 'G0005'] ],
                                                column_names=[ 'compoundnames', 'prop_preds' ], 
                                                figname='./foo--prop_masscenter.png',
                                                ):
    prop_means = []
    for i in range(len(filenames)):
        df = pd.read_csv(filenames[i], 
                         usecols=column_names
                         #names=['new_compoundnames', 'new_mols', 'new_smiles', 'new_prop_preds']
                         )
        
        prop_mean = df['prop_preds'].mean()
        print( i, prop_mean )
        
        prop_means.append( prop_mean )


    fig1, ax1 = plt.subplots(1, 1, figsize=(12, 8), sharey=False)
    
    plt.plot(range(len(prop_means)),
             prop_means,
             linestyle='-',
             color='r', 
             alpha=0.95, 
             label='prop_masscenter'
            )
    
    plt.legend(ncol=1, fontsize='x-large', shadow=True)
    ax1.grid(True)
    #fig1.show()
    plt.xlabel('generation' )
    plt.ylabel('prop_masscenter')
    plt.title('prop_masscenter vs generation')
    fig1.savefig(figname)

    
    return prop_means



def plot_abundance_anvCHs_vs_generation(filename='./abundance-vs-anvCHs.txt',
                                        sep=',',
                                        figname='./foo--abundance-anvCHs-vs-generation.png'
                                        ):
    df = pd.read_csv(filename, 
                     sep=sep,
                     #usecols=[ 'generation', 'abundance', 'anvCHs' ],
                     #names=[ 'generation', 'abundance', 'anvCHs' ],
                     )
    
    print("df.head()\n", df.head())
    print("df.columns:\n", df.columns)
    
    colnames = df.columns.tolist()
    colnames = colnames[1:]
    
    #print("colnames :\n", colnames)
    
    fig1, axs = plt.subplots(1, int(len(colnames)/2), figsize=(20, 6), sharey=False )
    
    for i in range( int(len(colnames)/2) ):
        df.plot(
                x='generation', 
                y=colnames[2*i], 
                title='%s and %s vs generation'%(colnames[2*i], colnames[2*i+1]),
                #kind='scatter', 
                kind='line', 
                linestyle='-',            #  linestyle=   '-',  '--', '-.', ':' 
                color='r', 
                alpha=0.95, 
                xlabel='generation',
                #ylabel='%s'%colnames[2*i],
                ax=axs[i]
                )
        
        axs[i].set_xlim([-0.5, 10.5])
        axs[i].set_ylim([-0.05, 1.05])
        axs[i].legend(loc='upper left')
        axs[i].grid(True)
        
        
        axs_i_twinx = axs[i].twinx()
        
        df.plot(
                x='generation', 
                y=colnames[2*i+1], 
                #title='%s vs generation'%colnames[2*i+1],
                #kind='scatter', 
                kind='line', 
                linestyle='-',
                color='b', 
                alpha=0.95, 
                #xlabel='generation',
                #ylabel='%s'%colnames[2*i+1],
                #ax=axs[i], 
                ax=axs_i_twinx
                )
        
        axs_i_twinx.set_ylim([0,20])
        axs_i_twinx.legend(loc='upper right')
        
    
    #axs.legend(loc='best')
    #axs.grid(True)
    #fig1.show()
    fig1.savefig(figname)



def compute_anvCHs_derivative_vs_generation(filename='./abundance-vs-anvCHs.txt',
                                            sep=',',
                                            xcolname='generation',
                                            ycolname='anvCHs1',
                                            ):
    df = pd.read_csv(filename, 
                     sep=sep,
                     )
    
    # print("df.head()\n", df.head())
    # print("df.columns:\n", df.columns)
    
    xs = df[xcolname].tolist()
    ys = df[ycolname].tolist()

    ds = []
    ds.append(np.NaN)

    for i in range(len(xs) - 1):    
        d = ( ys[i+1] - ys[i] )/ ( xs[i+1] - xs[i] )
        ds.append(d)

    ds[0] = ds[1]
    
    print("generation, anvCHs, anvCHs_derivative")
    for x, y, d in zip(xs, ys, ds):
        print("%d %10.4f %10.4f" %(x, y, d) )
    
    
    return xs, ys, ds
    
    
def plot_anvCHs_derivative_vs_generation(filename='./abundance-vs-anvCHs.txt',
                                         sep=',',
                                         ycolnames=['anvCHs1', 'anvCHs2', 'anvCHs3'],
                                         figname='./foo--anvCHs-derivative-vs-generation.png'
                                         ):

    xs, ys1, ds1 = compute_anvCHs_derivative_vs_generation(filename='./abundance-vs-anvCHs.txt',
                                                            sep=sep, 
                                                            xcolname='generation',
                                                            ycolname=ycolnames[0],
                                                            )
        
    xs, ys2, ds2 = compute_anvCHs_derivative_vs_generation(filename='./abundance-vs-anvCHs.txt',
                                                            sep=sep, 
                                                            xcolname='generation',
                                                            ycolname=ycolnames[1],
                                                            )

    xs, ys3, ds3 = compute_anvCHs_derivative_vs_generation(filename='./abundance-vs-anvCHs.txt',
                                                            sep=sep, 
                                                            xcolname='generation',
                                                            ycolname=ycolnames[2],
                                                            )

    new_df = pd.DataFrame(data=zip(xs, ds1, ds2, ds3), 
                          columns=(['generation', 'anvCHs1', 'anvCHs2', 'anvCHs3']))
    
    fig1, axs = plt.subplots(1, 3, figsize=(20, 6), sharey=False )
    
    for i, colname in zip( range(3), ['anvCHs1', 'anvCHs2', 'anvCHs3']) :
        new_df.plot(
                x='generation', 
                y=colname, 
                title='anvCHs_derivative vs generation',
                #kind='scatter', 
                kind='line', 
                linestyle='--',            #  linestyle=   '-',  '--', '-.', ':' 
                color='b', 
                alpha=0.95, 
                xlabel='generation',
                #ylabel='%s'%colnames[2*i],
                ax=axs[i]
                )
        
        axs[i].set_xlim([-0.5, 10.5])
        axs[i].set_ylim([-4.5, 0.5])
        axs[i].legend(loc='best')
        axs[i].grid(True)
    
    fig1.savefig(figname)


def compute_abundance_derivative_vs_generation(filename='./abundance-vs-anvCHs.txt',
                                                sep=',',
                                                xcolname='generation',
                                                ycolname='abundance',
                                                ):
    df = pd.read_csv(filename, 
                     sep=sep,
                     )
    
    # print("df.head()\n", df.head())
    # print("df.columns:\n", df.columns)
    
    xs = df[xcolname].tolist()
    ys = df[ycolname].tolist()

    ds = []
    ds.append(np.NaN)

    for i in range(len(xs) - 1):    
        d = ( ys[i+1] - ys[i] )/ ( xs[i+1] - xs[i] )
        ds.append(d)

    ds[0] = ds[1]
    
    print("generation, anvCHs1, anvCHs1_derivative")
    for x, y, d in zip(xs, ys, ds):
        print("%d %10.4f %10.4f" %(x, y, d) )
    
    
    return xs, ys, ds
    
    
def plot_abundance_derivative_vs_generation(filename='./abundance-vs-anvCHs.txt',
                                            sep=',', 
                                            ycolnames=['abundance1', 'abundance2', 'abundance3'],
                                            figname='./foo--abundance-derivative-vs-generation.png'
                                            ):

    xs, ys1, ds1 = compute_anvCHs_derivative_vs_generation(filename='./abundance-vs-anvCHs.txt',
                                                            sep=sep, 
                                                            xcolname='generation',
                                                            ycolname=ycolnames[0],
                                                            )
        
    xs, ys2, ds2 = compute_anvCHs_derivative_vs_generation(filename='./abundance-vs-anvCHs.txt',
                                                            sep=sep, 
                                                            xcolname='generation',
                                                            ycolname=ycolnames[1],
                                                            )

    xs, ys3, ds3 = compute_anvCHs_derivative_vs_generation(filename='./abundance-vs-anvCHs.txt',
                                                            sep=sep, 
                                                            xcolname='generation',
                                                            ycolname=ycolnames[2],
                                                            )

    new_df = pd.DataFrame(data=zip(xs, ds1, ds2, ds3), 
                          columns=(['generation', 'abundance1_deriv', 'abundance2_deriv', 'abundance3_deriv']))
    
    fig1, axs = plt.subplots(1, 3, figsize=(20, 6), sharey=False )
    
    for i, colname in zip( range(3), ['abundance1_deriv', 'abundance2_deriv', 'abundance3_deriv']) :
        new_df.plot(
                x='generation', 
                y=colname, 
                title='abundance_derivative vs generation',
                #kind='scatter', 
                kind='line', 
                linestyle='--',            #  linestyle=   '-',  '--', '-.', ':' 
                color='r', 
                alpha=0.95, 
                xlabel='generation',
                #ylabel='%s'%colnames[2*i],
                ax=axs[i]
                )
        
        axs[i].set_xlim([-0.5, 10.5])
        #axs[i].set_ylim([-4.5, 0.5])
        axs[i].legend(loc='best')
        axs[i].grid(True)
    
    fig1.savefig(figname)


#-----------------------------------------------------------------------------------------------
def accumulate_optmols(mutation_generation=1, 
                       old_acc_optmols_smi_path='./project/G%04d/acc_optmols.smi'%( 0 ),
                       new_mols_path='./project/G%04d/G%04d_df_ECFP.csv'%( 1, 1 ),
                       acc_optmols_smi_path='./project/G%04d/acc_optmols.smi'%( 1 ), 
                       column_names=[ 'smiles', 'prop_preds' ], 
                       critical_value=0.15,
                       is_write_sdf=False, 
                       outdir_path='./project/G%04d/acc_optmols/'%( 1 )
                       ):
    '''
    This function can accumulate the optimal mols by calculating the union of two sets of optmols (items are SMILES), 
    acc_optmols = old_acc_optmols.union( new_mols )
    The resultant acc_optmols smi file is written to acc_optmols.smi, 
    if desired (set is_write_sdf=True), the sdf file is also written to acc_optmols.sdf.
    

    Parameters
    ----------
    mutation_generation : int, optional
        DESCRIPTION. The default is 1. mutation_generation must be >= 0  integer,
                     if mutation_generation = 0, the acc_optmols = new_mols
    old_acc_optmols_smi_path : TYPE, optional
        DESCRIPTION. The default is './project/G%04d/acc_optmols.smi'%( 0 ).
    new_mols_path : TYPE, optional
        DESCRIPTION. The default is './project/G%04d/G%04d_df_ECFP.csv'%( 1, 1 ).
    acc_optmols_smi_path : TYPE, optional
        DESCRIPTION. The default is './project/G%04d/acc_optmols.smi'%( 1 ).
    column_names : TYPE, optional
        DESCRIPTION. The default is [ 'smiles', 'prop_preds' ].
    critical_value : TYPE, optional
        DESCRIPTION. The default is 0.15.
    is_write_sdf : TYPE, optional
        DESCRIPTION. The default is False.
    outdir_path : TYPE, optional
        DESCRIPTION. The default is './project/G%04d/acc_optmols/'%( 1 ).

    Returns
    -------
    TYPE
        DESCRIPTION.

    '''

    df = pd.read_csv(new_mols_path, 
                     usecols=column_names
                     )
    
    print( "\ndf.head()" )
    print( df.head() )
    print( "\ndf.describe()" )
    print( df.describe() )
    print( "\ndf.columns" )
    print( df.columns )
    
    try:
        idx_max = df.loc[df['prop_preds'].le(critical_value), 'prop_preds'].idxmax()
        print("idx_max = ", idx_max)
    except Exception as e:
        print(e)
        print("No mols exist whose prop_preds satisfy the condition: ", "prop_preds  <= ", critical_value)
        print("Set idx_max =  0")
        idx_max =  0
        

    if idx_max > 0:
        new_optmols_smis = df['smiles'].tolist()[ : idx_max ]
    else:
        print("!!! Warning, new_optmols_smis is an empty set.")
        new_optmols_smis = []
        
    new_optmols_smis = set( new_optmols_smis )
    print("\nlength of new_optmols_smis is : ", len(new_optmols_smis))

    if mutation_generation > 0:
        print("mutation_generation = %4d > 0"%mutation_generation )
    
        old_acc_optmols_suppl = Chem.SmilesMolSupplier( old_acc_optmols_smi_path, titleLine=False)
        # print("length of old_acc_optmols_suppl is : ", len(old_acc_optmols_suppl))
        old_acc_optmols = [x for x in old_acc_optmols_suppl if x is not None]
        # print("length of old_acc_optmols is : ", len(old_acc_optmols))
        old_acc_optmols_smis =  [ Chem.MolToSmiles( mol ) for mol in old_acc_optmols ]
        old_acc_optmols_smis = set(old_acc_optmols_smis)

    else:
        print("mutation_generation = %4d = 0"%mutation_generation )
        print("set old_acc_optmols_smis as empty set.")    
        old_acc_optmols_smis = set( )
        
    print("length of old_acc_optmols_smis is : ", len(old_acc_optmols_smis))    
    
    acc_optmols_smis = old_acc_optmols_smis.union( new_optmols_smis )
    acc_optmols_smis = list( acc_optmols_smis )
    acc_optmols_smis.sort()
    
    print("length of acc_optmols_smis is : ", len(acc_optmols_smis))

    with open(acc_optmols_smi_path, 'w') as fr:
        for smi in acc_optmols_smis:
            fr.write(smi + '\n')

    
    if is_write_sdf == True:
        #---- write molecules to SD File -------------------------------        
        print("\nWrite to sdf file.\n")
        
        All_mols = [ Chem.MolFromSmiles( smi ) for smi in acc_optmols_smis if Chem.MolFromSmiles(smi) is not None]
        
        cpd_names = [ "cpd-%05d" % (i+1) for i in range(len(All_mols)) ]
        for mol, cpd_name in zip(All_mols, cpd_names):
            mol.SetProp("_Name", cpd_name )

        call_mkdir_acc_optmols = 'mkdir   ' +  outdir_path  +  '    2>>/dev/null  '
        status, output = subprocess.getstatusoutput( call_mkdir_acc_optmols )
        print("mkdir  ./project/GXXXX/acc_optmols;    status, output = ", status, output)        

        print("call write_mols_paralell(), the separate acc_optmols are written to dir:  %s"%outdir_path )
        write_mols_paralell(All_mols, cpd_names, outfile_path=outdir_path)

        print("cat cpd*.sdf to acc_optmols.sdf")
        status, output = subprocess.getstatusoutput( 'rm  -f   ./project/' +  'G%04d'%mutation_generation  +  '/acc_optmols.sdf   2>>/dev/null ' )
        print("status, output = ", status, output)
        cat_sdfs_to_onefile = 'cat   '  +  outdir_path  + 'cpd*.sdf  >>  '  +  './project/' + 'G%04d'%mutation_generation  + '/acc_optmols.sdf   '
        status, output = subprocess.getstatusoutput( cat_sdfs_to_onefile )
        print("status, output = ", status, output)

    else:
        print("\nDo not write to sdf file.\n")

   
    return len( acc_optmols_smis )



def accumulate_optmols_S1En(mutation_generation=1, 
                            old_acc_optmols_smi_paths=[ './project/G%04d/acc_optmols--S1En-%s.smi'%( 0, color ) for color in ["R", "G", "B"] ],
                            new_mols_path='./project/G%04d/G%04d_df_ECFP--S1En.csv'%( 1, 1 ),
                            acc_optmols_smi_paths=[ './project/G%04d/acc_optmols--S1En-%s.smi'%( 1, color ) for color in ["R", "G", "B"] ],
                            column_names=[ 'compoundnames', 'smiles', 'prop_preds' ], 
                            critical_values=[ 2.70, 3.00 ],
                            ):
    '''
    This function can accumulate the optimal mols by calculating the union of two sets of optmols (items are SMILES), 
    acc_optmols = old_acc_optmols.union( new_mols )
    The resultant acc_optmols smi file is written to acc_optmols.smi, 
    if desired (set is_write_sdf=True), the sdf file is also written to acc_optmols.sdf.
    

    Parameters
    ----------
    mutation_generation : int, optional
        DESCRIPTION. The default is 1. mutation_generation must be >= 0  integer,
                     if mutation_generation = 0, the acc_optmols = new_mols
    old_acc_optmols_smi_path : TYPE, optional
        DESCRIPTION. The default is './project/G%04d/acc_optmols--S1En.smi'%( 0 ).
    new_mols_path : TYPE, optional
        DESCRIPTION. The default is './project/G%04d/G%04d_df_ECFP--S1En.csv'%( 1, 1 ).
    acc_optmols_smi_path : TYPE, optional
        DESCRIPTION. The default is './project/G%04d/acc_optmols--S1En.smi'%( 1 ).
    column_names : TYPE, optional
        DESCRIPTION. The default is [ 'smiles', 'prop_preds' ].
    critical_values : TYPE, optional
        DESCRIPTION. The default is [2.70, 3.00].

    Returns
    -------
    TYPE
        DESCRIPTION.

    '''

    print("Call accumulate_optmols_S1En():\n")    

    df = pd.read_csv(new_mols_path, 
                     usecols=column_names
                     )
    
    print( "\ndf.head()" )
    print( df.head() )
    print( "\ndf.describe()" )
    print( df.describe() )
    print( "\ndf.columns" )
    print( df.columns )

    #--- define  gt_idxmax_B for blue color. energy > ENERGY_THRESHOLD2 = 2.80 ev -------
    #--- if gt_idxmax_B > 0, the indexs for blue mols are  [0 : gt_idmax_B]
    #--- then, le_idxmin_G is the ending index for green color. 
    lt_idxmin_G, le_idxmin_G, gt_idxmax_B, ge_idxmax_B = getmaxminidxs_lt_le_gt_ge_criticval_df(df, 
                                                                                                colname='prop_preds',
                                                                                                critical_value=critical_values[1],
                                                                                                ascending=False)
    
    #--- define  lt_idxmin_R for red color. energy < ENERGY_THRESHOLD1 = 2.30 ev -------
    #--- if lt_idxmin_R > 0, the indexs for red mols are  [lt_idmin_R : ]
    #--- then, ge_idxmax_G is the starting index for green color. 
    lt_idxmin_R, le_idxmin_R, gt_idxmax_G, ge_idxmax_G = getmaxminidxs_lt_le_gt_ge_criticval_df(df, 
                                                                                                colname='prop_preds',
                                                                                                critical_value=critical_values[0],
                                                                                                ascending=False)
    
    if (gt_idxmax_B > 0):
        optimal_smis_B = df.smiles.tolist()[0 : gt_idxmax_B]
        optimal_cpdnames_B = df.compoundnames.tolist()[0 : gt_idxmax_B]
        prop_preds_S1En_B = df.prop_preds.tolist()[0 : gt_idxmax_B]
        print("-"*60)
        print("-#-Blue mols, the optimal_cpdname, optimal_smis and prop_preds_S1En are:")
        for cpdname, smi, S1En in zip(optimal_cpdnames_B, optimal_smis_B, prop_preds_S1En_B):
            print(cpdname, " , ", smi, " , %8.4f"%S1En )    
    else:
        optimal_smis_B = set()
        print("-#-Blue mols is empty!")
    
    
    if (le_idxmin_G < df.shape[0]) and (ge_idxmax_G < df.shape[0]) and (le_idxmin_G < ge_idxmax_G):
        optimal_smis_G = df.smiles.tolist()[le_idxmin_G : ge_idxmax_G]
        optimal_cpdnames_G = df.compoundnames.tolist()[le_idxmin_G : ge_idxmax_G]
        prop_preds_S1En_G = df.prop_preds.tolist()[le_idxmin_G : ge_idxmax_G]
        print("-"*60)
        print("-#-Green mols, the optimal_cpdname, optimal_smis and prop_preds_S1En are:")
        for cpdname, smi, S1En in zip(optimal_cpdnames_G, optimal_smis_G, prop_preds_S1En_G):
            print(cpdname, " , ", smi, " , %8.4f"%S1En )    
    else:
        optimal_smis_G = set()
        print("-#-Green mols is empty!")
    
    
    if (lt_idxmin_R < df.shape[0]):
        optimal_smis_R = df.smiles.tolist()[lt_idxmin_R : ]
        optimal_cpdnames_R = df.compoundnames.tolist()[lt_idxmin_R : ]
        prop_preds_S1En_R = df.prop_preds.tolist()[lt_idxmin_R : ]
        print("-"*60)
        print("-#-Red mols, the optimal_cpdname, optimal_smis and prop_preds_S1En are:")
        for cpdname, smi, S1En in zip(optimal_cpdnames_R, optimal_smis_R, prop_preds_S1En_R):
            print(cpdname, " , ", smi, " , %8.4f"%S1En )    
    else:
        optimal_smis_R = set()
        print("-#-Red mols is empty!")
    
    
    print("-#-Size of Blue, Green, Red mols are: ", len(optimal_smis_B), len(optimal_smis_G), len(optimal_smis_R))


    new_optmols_smis_all = [ optimal_smis_R, optimal_smis_G, optimal_smis_B ]
    acc_optmols_smis_all = []
    for  i in range(3):
        print("-#- ", i)
        new_optmols_smis = set( new_optmols_smis_all[i] )
        print("\nlength of new_optmols_smis_S1En is : ", len(new_optmols_smis))
    
        if mutation_generation > 0:
            print("mutation_generation = %4d > 0"%mutation_generation )
            
            try:
                old_acc_optmols_suppl = Chem.SmilesMolSupplier( old_acc_optmols_smi_paths[i], titleLine=False)
                old_acc_optmols = [x for x in old_acc_optmols_suppl if x is not None]
                old_acc_optmols_smis =  [ Chem.MolToSmiles( mol ) for mol in old_acc_optmols ]
                old_acc_optmols_smis = set(old_acc_optmols_smis)
            except Exception as e:
                print(e)
                print("*** Set old_acc_optmols_smis an empty set.")
                old_acc_optmols_smis = set()
    
        else:
            print("mutation_generation = %4d "%mutation_generation )
            print("set old_acc_optmols_smis_S1En as empty set.")    
            old_acc_optmols_smis = set( )
            
        print("length of old_acc_optmols_smis_S1En is : ", len(old_acc_optmols_smis))    
        
        acc_optmols_smis = old_acc_optmols_smis.union( new_optmols_smis )
        acc_optmols_smis = list( acc_optmols_smis )
        acc_optmols_smis.sort()
        
        print("length of acc_optmols_smis_S1En is : ", len(acc_optmols_smis))
    
        with open(acc_optmols_smi_paths[i], 'w') as fr:
            #fr.write("#smiles\n" )
            for smi in acc_optmols_smis:
                fr.write(smi + '\n')

        acc_optmols_smis_all.append(acc_optmols_smis)
   
    return len( acc_optmols_smis_all[0] ), len( acc_optmols_smis_all[1] ), len( acc_optmols_smis_all[2] )



def sort_mols_by_SAScores(mutation_generation=1,
                          acc_optmols_smi_path='./project/G0001/acc_optmols.smi', 
                          column_names=[ 'mols', 'smiles' ], 
                          outdir_path='./project/G0001/acc_optmols_sortedbysas/',
                          is_write_sdf=True, 
                          new_sdf_path='./project/G0001/acc_optmols_sortedbysas.sdf',
                          ascending=True
                          ):
    
    acc_optmols_suppl = Chem.SmilesMolSupplier( acc_optmols_smi_path, titleLine=False)
    print("length of acc_optmols_suppl is : ", len(acc_optmols_suppl))
    acc_optmols = [x for x in acc_optmols_suppl if x is not None]
    print("length of acc_optmols is : ", len(acc_optmols))
    acc_optmols_smis =  [ Chem.MolToSmiles( mol ) for mol in acc_optmols ]
    
    
    df = pd.DataFrame(data=zip(acc_optmols, acc_optmols_smis), 
                          columns=column_names
                          )

    df['SAscores'] = df.mols.map(sascorer.calculateScore)

    print( "\ndf.columns" )
    print( df.columns )
    print( "\ndf.describe()" )
    print( df.describe() )    
    # print( "\ndf.head()" )
    # print( df.head() )
    
    # (id_max, id_min) = (df.calc_SA_score.idxmax(), df.calc_SA_score.idxmin())

    df.sort_values(by="SAscores", ascending=ascending, inplace=True)

    print('# The optimal 16 molecules are: ')
    print( df.head(16) )

    if (is_write_sdf == True):
        #---- write molecules to SD File -------------------------------        
        print("\nWrite to sdf file.\n")
        
        All_mols = df['mols'].tolist()
        
        cpd_names = [ "cpd-%05d" % (i+1) for i in range(len(All_mols)) ]
        for mol, cpd_name in zip(All_mols, cpd_names):
            mol.SetProp("_Name", cpd_name )

        call_mkdir_acc_optmols_sorted = 'mkdir   ' +  outdir_path  +  '    2>/dev/null  '
        status, output = subprocess.getstatusoutput( call_mkdir_acc_optmols_sorted )
        print("mkdir  ./project/GXXXX/acc_optmols_sortedbysas/;    status, output = ", status, output)        

        print("call write_mols_paralell(), the separate acc_optmols are written to dir:  %s"%outdir_path )
        write_mols_paralell(All_mols, cpd_names, outfile_path=outdir_path)

        print("cat cpd*.sdf to acc_optmols.sdf")
        status, output = subprocess.getstatusoutput( 'rm  -f   ./project/' +  'G%04d'%mutation_generation  +  '/acc_optmols_sortedbysas.sdf   2>/dev/null ' )
        print("status, output = ", status, output)
        cat_sdfs_to_onefile = 'cat   '  +  outdir_path  + 'cpd*.sdf  >>  '  +  './project/' + 'G%04d'%mutation_generation  + '/acc_optmols_sortedbysas.sdf   '
        status, output = subprocess.getstatusoutput( cat_sdfs_to_onefile )
        print("status, output = ", status, output)

    else:
        print("\nDo not write to sdf file.\n")


    return df



def accumulate_optmols_intersection(mutation_generation=0, 
                                    acc_optmols_smi_path1='./project/G%04d/acc_optmols.smi'%( 0 ),
                                    acc_optmols_smi_path2='./project/G%04d/acc_optmols--S1En-B.smi'%( 0 ),
                                    acc_optmols_smi_path3='./project/G%04d/acc_optmols--INTERSCTION-B.smi'%( 0 ), 
                                    is_write_sdf=False, 
                                    outdir_path='./project/G%04d/acc_optmols--INTERSCTION-B/'%( 0 )
                                    ):


    acc_optmols_suppl1 = Chem.SmilesMolSupplier( acc_optmols_smi_path1, titleLine=False)
    acc_optmols1 = [x for x in acc_optmols_suppl1 if x is not None]
    acc_optmols_smis1 =  [ Chem.MolToSmiles( mol ) for mol in acc_optmols1 ]
    acc_optmols_smis1 = set(acc_optmols_smis1)
        
    print("length of acc_optmols_smis1 is : ", len(acc_optmols_smis1))    
    
    acc_optmols_suppl2 = Chem.SmilesMolSupplier( acc_optmols_smi_path2, titleLine=False)
    acc_optmols2 = [x for x in acc_optmols_suppl2 if x is not None]
    acc_optmols_smis2 =  [ Chem.MolToSmiles( mol ) for mol in acc_optmols2 ]
    acc_optmols_smis2 = set(acc_optmols_smis2)
    
    print("length of acc_optmols_smis2 is : ", len(acc_optmols_smis2))    
    
    acc_optmols_smis = acc_optmols_smis1.intersection( acc_optmols_smis2 )
    acc_optmols_smis = list( acc_optmols_smis )
    acc_optmols_smis.sort()
    
    print("length of acc_optmols_smis (the intersection set) is : ", len(acc_optmols_smis))

    with open(acc_optmols_smi_path3, 'w') as fr:
        for smi in acc_optmols_smis:
            fr.write(smi + '\n')

    
    if is_write_sdf == True:
        #---- write molecules to SD File -------------------------------        
        print("\nWrite to sdf file.\n")
        
        All_mols = [ Chem.MolFromSmiles( smi ) for smi in acc_optmols_smis if Chem.MolFromSmiles(smi) is not None]
        
        cpd_names = [ "cpd-%05d" % (i+1) for i in range(len(All_mols)) ]
        for mol, cpd_name in zip(All_mols, cpd_names):
            mol.SetProp("_Name", cpd_name )

        call_mkdir_acc_optmols = 'mkdir   ' +  outdir_path  +  '    2>>/dev/null  '
        status, output = subprocess.getstatusoutput( call_mkdir_acc_optmols )
        print("mkdir  ./project/GXXXX/acc_optmols--INTERSCTION-B;    status, output = ", status, output)        

        print("call write_mols_paralell(), the separate acc_optmols--INTERSCTION-B are written to dir:  %s"%outdir_path )
        write_mols_paralell(All_mols, cpd_names, outfile_path=outdir_path)

        print("cat cpd*.sdf to acc_optmols--INTERSCTION-B.sdf")
        status, output = subprocess.getstatusoutput( 'rm  -f   ./project/' +  'G%04d'%mutation_generation  +  '/acc_optmols--INTERSCTION-B.sdf   2>>/dev/null ' )
        print("status, output = ", status, output)
        cat_sdfs_to_onefile = 'cat   '  +  outdir_path  + 'cpd*.sdf  >>  '  +  './project/' + 'G%04d'%mutation_generation  + '/acc_optmols--INTERSCTION-B.sdf   '
        status, output = subprocess.getstatusoutput( cat_sdfs_to_onefile )
        print("status, output = ", status, output)

    else:
        print("\nDo not write to sdf file.\n")

   
    return len( acc_optmols_smis )




#----------------------- maps by FPSimilarity ----------------------------------
def get_intersection_plus_smiles(smis1=['CCOC', 'CCO', 'COC', 'COOC'], 
                                 smis2=['CNCCOC', 'CCOOC', 'CNCCO', 'CNCOC']
                                 ):
    #--- assume size of list1 <= list2 ------------------------------
    #--- further assume elements in list are unique. ------------------------
    print('-'*60)
    
    exchange_flag = False
    if len(smis1) > len(smis2):
        print("len(smis1) > len(smis2), exchange each other.")
        smis1, smis2 = smis2, smis1
        exchange_flag = True
    
    smis1_set = set(smis1)
    smis2_set = set(smis2)
    
    intersection_smis_set = smis1_set.intersection(smis2_set)
    intersection_smis = list(intersection_smis_set)
    intersection_smis.sort()
    
    smis1_dif_smis2_set = smis1_set.difference(smis2_set)
    smis1_dif_smis2 = list(smis1_dif_smis2_set)
    smis1_dif_smis2.sort()
    
    smis2_dif_smis1_set = smis2_set.difference(smis1_set)
    smis2_dif_smis1 = list(smis2_dif_smis1_set)
    smis2_dif_smis1.sort()
    
    print("len(smis1) = ", len(smis1) )
    print("len(smis2) = ", len(smis2) )
    print("len(set(smis1()) = ", len(set(smis1)) )
    print("len(set(smis2)) = ", len(set(smis2)) )
    print("len(intersection_smis) = ", len(intersection_smis) )
    print("len(smis1_dif_smis2) = ", len(smis1_dif_smis2) )
    print("len(smis2_dif_smis1) = ", len(smis2_dif_smis1) )
    # print("len(smis1_dif_smis2 + intersection_smis) = ", len(smis1_dif_smis2 + intersection_smis) )
    # print("len(smis2_dif_smis1 + intersection_smis) = ", len(smis2_dif_smis1 + intersection_smis) )
    
    print("\nsmis1:\n", smis1)
    print("smis2:\n", smis2)
    print("intersection_smis:\n", intersection_smis)
    print("smis1_dif_smis2:\n", smis1_dif_smis2)
    print("smis2_dif_smis1:\n", smis2_dif_smis1)
    
    smis1_rev  =  intersection_smis + smis1_dif_smis2
    smis2_rev  =  intersection_smis + smis2_dif_smis1
    
    return intersection_smis, smis1_rev, smis2_rev, exchange_flag



def compute_similarity_matrix(smis1=['CCOC', 'CCO'], 
                              smis2=['CNCCOC', 'CCOOC', 'CNCCO', 'CNCOC'],
                              fpSize=2048
                              ):
    #--- assume number of rows <= number of columns ------------------------------
    print('-'*60)
    
    if len(smis1) > len(smis2):
        print("exchange smis1, smis2")
        smis1, smis2 = smis2, smis1
    
    print("smis1:\n", smis1)
    print("smis2:\n", smis2)
    
    ms1 = [Chem.MolFromSmiles(smi) for smi in smis1]
    ms2 = [Chem.MolFromSmiles(smi) for smi in smis2]
    
    #Chem.RDKFingerprint(x, fpSize=2048)  #  Returns an RDKit topological fingerprint for a molecule
    fps1 = [Chem.RDKFingerprint(x, fpSize=fpSize) for x in ms1]
    fps2 = [Chem.RDKFingerprint(x, fpSize=fpSize) for x in ms2]
    
    n1 = len(fps1)
    n2 = len(fps2)
    
    arr_sim = np.zeros(shape=(n1, n2), 
                       dtype=np.float64
                       )  
    
    for i in range(n1):
        for j in range(n2):
            tempt_sim = DataStructs.FingerprintSimilarity(fps1[i], fps2[j])
            arr_sim[i][j] = tempt_sim
    
    mat_sim = np.matrix(data=arr_sim,
                        dtype=arr_sim.dtype)
    
    print("\nmat_sim:\n", mat_sim)
    
    return mat_sim



def map_molslibs_by_similarity(smis1=['CCOC', 'CCO'], 
                               smis2=['CNCCOC', 'CCOOC', 'CNCCO', 'CNCOC'],
                               fpSize=2048,
                               is_return_max_app_sim=False
                               ):
    #--- assume number of rows <= number of columns ------------------------------
    #--- assume size of list1 <= list2 ------------------------------
    #--- further assume elements in list are unique. ------------------------
    #------------------------------------------------------------------------------
    # A[[i, j], :] = A[[j, i], :] # 实现了第i行与第j行的互换
    #-----------------------------------------------------------------------------
    if len(smis1) > len(smis2):
        print("len(smis1) > len(smis2), exchange each other.")
        smis1, smis2 = smis2, smis1
    
    n1 = len(smis1)
    n2 = len(smis2)
    
    mat_sim = compute_similarity_matrix(smis1=smis1, 
                                        smis2=smis2,
                                        fpSize=fpSize
                                        )
    
    ms1_indexs = list( range(n1) )
    ms2_indexs = list( range(n2) )

    #--- transform matrix by similarity maxium rule. --------------------------
    print('/'*60)
    print("\nBegin transform mat_sim.")
    for i in range(n1 - 1):
        print('-'*60)
        print("row (or column) index = %4d"%i)
        # print("\nmat_sim, original:\n", mat_sim)
        
        max_index = np.unravel_index(mat_sim[i:n1, i:n2].argmax(), mat_sim[i:n1, i:n2].shape)
            
        # print("\nmat_sim[i:n1, i:n2]:\n", mat_sim[i:n1, i:n2])
        # print("\nmax_index: \n", max_index)
        max_row_index = max_index[0] + i
        max_column_index = max_index[1] + i
        print("max_row_index = ", max_row_index, "; max_column_index = ", max_column_index )
        print("max_value = ", mat_sim[max_row_index, max_column_index])
    
        if (max_row_index != i):
            mat_sim[[i, max_row_index], :] = mat_sim[[max_row_index, i], :]
            # print("\nmat_sim, after row_exchange: %4d <--> %4d\n"%(i, max_row_index), mat_sim)
            
            # print("Before exchange row: ", ms1_indexs[i], ms1_indexs[max_row_index])
            tempt_row_index  = ms1_indexs[i]
            ms1_indexs[i] = ms1_indexs[max_row_index]
            ms1_indexs[max_row_index] = tempt_row_index
            # print("After exchange row: ", ms1_indexs[i], ms1_indexs[max_row_index])
    
        if (max_column_index != i):
            mat_sim[: , [i, max_column_index] ] = mat_sim[: , [max_column_index, i]]
            # print("\nmat_sim, after column_exchange: %4d <--> %4d\n"%(i, max_column_index), mat_sim)
            
            # print("Before exchange column: ", ms2_indexs[i], ms2_indexs[max_column_index])
            tempt_column_index  = ms2_indexs[i]
            ms2_indexs[i] = ms2_indexs[max_column_index]
            ms2_indexs[max_column_index] = tempt_column_index
            # print("After exchange column: ", ms2_indexs[i], ms2_indexs[max_column_index])
    
    print("\nEnd transform mat_sim.")
    print('\\'*60)
        
    new_smis1 = [smis1[idx] for idx in ms1_indexs]
    new_smis2 = [smis2[idx] for idx in ms2_indexs]

    print("Restrict the size of new_smis2 by the size of new_smis1.")
    new_smis2 = new_smis2[:len(new_smis1)]

    print('-'*60)
    print("\nms1_indexs:", ms1_indexs)
    print("ms2_indexs:", ms2_indexs)
    print("smis1:\n", smis1)
    print("smis2:\n", smis2)    
    print("new_smis1:\n", new_smis1)
    print("new_smis2:\n", new_smis2)
    print("\nmat_sim:\n", mat_sim)

    if (is_return_max_app_sim == False):
        return new_smis1, new_smis2, mat_sim
    else:
        diag_sim = np.diag(mat_sim)
        if (diag_sim.shape[0] > 0):
            max_app_sim = diag_sim.sum()/diag_sim.shape[0]
        else:
            print("diag_sim.shape[0] <= 0")
            exit(-1)
        print("diag_sim:\n", diag_sim)
        print("diag_sim.shape = ", diag_sim.shape)
        print("max_app_sim = ", max_app_sim)    
        
        return new_smis1, new_smis2, mat_sim, max_app_sim
        


def get_centroid_series(smis=[ 'CCCNCCO', 'CCNCCO', 'CC' ],
                        is_return_arranged_smis=False
                        ):
    
    mat_sim = compute_similarity_matrix(smis1=smis, 
                                        smis2=smis,
                                        fpSize=2048
                                        )

    n1 = len(smis)
    n2 = n1
    ms1_indexs = list( range(n1) )

    #--- transform matrix by similarity maximum rule. --------------------------
    print('/'*60)
    print("Compute a series of similarity mass centers.")
    print("\nBegin transform mat_sim.")
    print("mat_sim: \n", mat_sim)
    for i in range(n1 - 1):
        print('-'*60)
        print("row (or column) index = %4d"%i)
        print("exchange before, mat_sim[i:n1, i:n2] : \n", mat_sim[i:n1, i:n2])
        mat_sim_row_means = mat_sim[i:n1, i:n2].mean(axis=1)
        print("mat_sim_row_means:\n", mat_sim_row_means)
        row_max_idx = mat_sim_row_means.argmax() + i
        print("row_max_idx = ", row_max_idx)
        print("Centriod mol: ", smis[row_max_idx])
        
        if row_max_idx != i:
            mat_sim[[i, row_max_idx], :] = mat_sim[[row_max_idx, i], :]
            mat_sim[: , [i, row_max_idx] ] = mat_sim[: , [row_max_idx, i]]
            ms1_indexs[i], ms1_indexs[row_max_idx] = ms1_indexs[row_max_idx], ms1_indexs[i]
        
        print("exchange after, mat_sim[i:n1, i:n2] : \n", mat_sim[i:n1, i:n2])
        print("ms1_indexs:\n", ms1_indexs)
    
    print("\nEnd transform mat_sim.")
    
    centriod_smis = [smis[idx] for idx in ms1_indexs]
    print("\ncentriod_smis:\n", centriod_smis)

    #--------- is return arranged smis by shifting the middle ones to the middle of the series.
    if (is_return_arranged_smis == True):
        print('\\'*60)
        print("len(ms1_indexs) = ", len(ms1_indexs))
        print("len(ms1_indexs)/2 = ", int(len(ms1_indexs)/2) )
        print("ms1_indexs:\n", ms1_indexs)
        # mid_idx = int( len(ms1_indexs)/2 )
        new_middle_idxs = []
        
        mol_0 = Chem.MolFromSmiles( smis[ ms1_indexs[0] ])
        fp_0 = Chem.RDKFingerprint(mol_0, fpSize=2048)
        for i, j in zip(range(0, len(ms1_indexs), 2), range(1, len(ms1_indexs), 2)):
            print(i, j, " : ", ms1_indexs[i], ms1_indexs[j])
            mol_a = Chem.MolFromSmiles( smis[ ms1_indexs[i] ])
            mol_b = Chem.MolFromSmiles( smis[ ms1_indexs[j] ])
            fp_a = Chem.RDKFingerprint(mol_a, fpSize=2048)
            fp_b = Chem.RDKFingerprint(mol_b, fpSize=2048)
            sim_a0 = DataStructs.FingerprintSimilarity( fp_a,  fp_0)
            sim_b0 = DataStructs.FingerprintSimilarity( fp_b,  fp_0)
            print("sim_a0 = ", sim_a0)
            print("sim_b0 = ", sim_b0)
            if sim_a0 > sim_b0:
                ms1_indexs[i], ms1_indexs[j] = ms1_indexs[j], ms1_indexs[i]
                
            print("append, ", ms1_indexs[i], "; insert ", ms1_indexs[j])
            new_middle_idxs.append(ms1_indexs[i])
            new_middle_idxs.insert(0, ms1_indexs[j])
            print("new_middle_idxs:\n", new_middle_idxs)
        
        if len(ms1_indexs) % 2 == 1:
            print("append, ", ms1_indexs[-1])
            new_middle_idxs.append(ms1_indexs[-1])
            print("new_middle_idxs:\n", new_middle_idxs)
    
        print("Final new_middle_idxs:\n", new_middle_idxs)
        new_smis = [ smis[idx] for idx in new_middle_idxs]
        print("new_smis:\n", new_smis)
        
        mat_sim_final = compute_similarity_matrix(smis1=[ centriod_smis[0]  ], 
                                                  smis2=new_smis,
                                                  fpSize=2048
                                                  )
    
    else:

        mat_sim_final = compute_similarity_matrix(smis1=[ centriod_smis[0]  ], 
                                                  smis2=centriod_smis,
                                                  fpSize=2048
                                                  )

    
    mat_sim_row_means_final = mat_sim_final.mean(axis=1)
    print("mat_sim_row_means_final:\n", mat_sim_row_means_final)
    max_ave_sim = mat_sim_row_means_final[0]
    print("\nmax_nave_sim = %10.3f"%max_ave_sim)

    # mat_sim_row_means_global = mat_sim.mean(axis=1)
    # print("mat_sim_row_means_global:\n", mat_sim_row_means_global)
    # global_row_max_idx = mat_sim_row_means_global.argmax()
    # ave_sim = mat_sim_row_means_global[global_row_max_idx]
    # print("\nave_sim = %10.4f"%ave_sim)


    if (is_return_arranged_smis == False):
        return centriod_smis, max_ave_sim
    else:
        return centriod_smis, max_ave_sim, new_smis

    

def get_centroid(smis=[ 'CCCNCCO', 'CCNCCO', 'CC' ]
                 ):
    
    mat_sim = compute_similarity_matrix(smis1=smis, 
                                        smis2=smis,
                                        fpSize=2048
                                        )

    print("mat_sim: \n", mat_sim)
    mat_sim_row_means = mat_sim.mean(axis=1)
    print("mat_sim_row_means:\n", mat_sim_row_means)
    row_max_idx = mat_sim_row_means.argmax()
    print("row_max_idx = ", row_max_idx)
    
    centriod_smi = smis[row_max_idx]
    max_ave_sim = mat_sim_row_means[row_max_idx]
    
    print("\ncentriod_smi: \n", smis[row_max_idx])
    print("\nmax_ave_sim = %10.3f"%max_ave_sim)

    return centriod_smi, max_ave_sim
    
    
def draw_mols_and_show(smis,
                       figname='foo--mols.png',
                       molsPerRow=3,
                       subImgSize=(800, 800),
                       num_of_mols=None
                       ):
    #-----------------------------------------------------------------------------------------
    if (num_of_mols is None):
        mols1 = [Chem.MolFromSmiles(smi) for smi in smis]
    elif (num_of_mols > 0):
        mols1 = [Chem.MolFromSmiles(smi) for smi in smis[ : num_of_mols] ]
    else:
        print("parameter num_of_mols must be larger than 0 or equal to None. Check it.")
        exit(-1)
    
    img1 = Draw.MolsToGridImage(mols1, 
                                molsPerRow=molsPerRow, 
                                subImgSize=subImgSize, 
                                returnPNG=False)
    
    img1.save(figname)
 
    lena1 = mpimg.imread(figname)    # 读取和代码处于同一目录下的 lena.png  # 此时 lena 就已经是一个 np.array 了，可以对它进行任意处理
    # print("lena1.shape = ", lena1.shape)       #(512, 512, 3)
    plt.imshow(lena1) # 显示图片
    plt.axis('off') # 不显示坐标轴
    plt.show()
    


def load_mols_in_sdf(sdf_path='./data/Pick_mols.sdf'):
    #-------- loading  sdf file ---------------------------------------------------
    sdsuppl = Chem.SDMolSupplier(sdf_path)
    mols = [x for x in sdsuppl if x is not None]
    smiles = [Chem.MolToSmiles(mol) for mol in mols]
    print("-"*60)  
    print( "number of molecules is ", len(mols) )
    
    return mols, smiles


def plot_muation_diagram(filename='./project/mutation_diagram.csv',
                         figname='foo--muation-diagram.png',
                         molsPerRow=3,
                         subImgSize=(800, 800),
                         num_of_mols=None
                         ):
    
    df = pd.read_csv(filename)
    # mutation_generation, centriod_smi1, len_smis1, inner_max_ave_sim1, inter_max_app_sim
    
    print( "\ndf.head()" )
    print( df.head() )
    print( "\ndf.describe()" )
    print( df.describe() )
    print( "\ndf.columns" )
    print( df.columns )    
    
    # mutation_generations = df['mutation_generation'].tolist()
    centriod_smis = df['centriod_smi1'].tolist()
    len_smis = df['len_smis1'].tolist()
    inner_max_ave_sims = df['inner_max_ave_sim1'].tolist()
    inter_max_app_sims = df['inter_max_app_sim'].tolist()
    
    legends_ = [ "(%5d, %6.3f, %6.3f)"%(ele1, ele2, ele3) for ele1, ele2, ele3 in zip(len_smis, inner_max_ave_sims, inter_max_app_sims) ]
    
    #-----------------------------------------------------------------------------------------
    if (num_of_mols is None):
        mols1 = [Chem.MolFromSmiles(smi) for smi in centriod_smis]
    elif (num_of_mols > 0):
        mols1 = [Chem.MolFromSmiles(smi) for smi in centriod_smis[ : num_of_mols] ]
    else:
        print("parameter num_of_mols must be larger than 0 or equal to None. Check it.")
        exit(-1)
    
    img1 = Draw.MolsToGridImage(mols1, 
                                molsPerRow=molsPerRow, 
                                subImgSize=subImgSize, 
                                legends=legends_,
                                returnPNG=False)
    
    img1.save(figname)
 
    lena1 = mpimg.imread(figname)    # 读取和代码处于同一目录下的 lena.png  # 此时 lena 就已经是一个 np.array 了，可以对它进行任意处理
    # print("lena1.shape = ", lena1.shape)       #(512, 512, 3)
    plt.imshow(lena1) # 显示图片
    plt.axis('off') # 不显示坐标轴
    plt.show()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 14 11:18:02 2021

@author: tucy
"""

import sys 
import os
#os.chdir('../')
script_wkdir = os.getcwd()
sys.path.append(script_wkdir)

from salam.module_MD.D_Pi_A_Enumeration import draw_mols
from rdkit import Chem
#from rdkit.Chem import AllChem

from tqdm import tqdm
from salam.module_MD.D_Pi_A_Enumeration import try_embedmolecule
from salam.module_MD.D_Pi_A_Enumeration import delete_repeated_smiles_of_mols

#from salam.module_MD.OrganoMetalics_Enumeration import mutations_CH_aromatic
#from salam.module_MD.OrganoMetalics_Enumeration import mutations_CH_aromatic_termino_H
from salam.module_MD.OrganoMetalics_Enumeration import  debranch_CR_aromatic
from salam.module_MD.OrganoMetalics_Enumeration import  mutations_N_to_CH_aromatic


import subprocess

import numpy as np
#from scipy.special import comb

import argparse

parser = argparse.ArgumentParser(description="manual to this script:")
# parser.add_argument('--N_PATT', type=str, default='[n&R&H0,N&R&H0,N&R&H1]' )
# parser.add_argument('--CR_PATT', type=str, default='[a&R]-[A&!#1,a&!#1]' )

parser.add_argument('--N_PATT', type=str, default='[n&R&H0]' )
parser.add_argument('--CR_PATT', type=str, default='[c&R]-[A&!#1]' )

parser.add_argument('--GXXXX', type=str, default="G0000")
parser.add_argument('--PXXXX', type=str, default="P0000")
parser.add_argument('--OXXXX', type=str, default="O0000")
parser.add_argument('--GXXXXPLUSONE', type=str, default="G0001")

args = parser.parse_args()

print("args.N_PATT: ", args.N_PATT, type(args.N_PATT))
N_PATT = args.N_PATT

print("args.CR_PATT: ", args.CR_PATT, type(args.CR_PATT))
CR_PATT = args.CR_PATT

#TERMINO_HYDROGEN_SUBS  = [ ele.strip()  for ele in TERMINO_HYDROGEN_SUBS.strip().split(",") ]


print(args.GXXXX, type(args.GXXXX))
print(args.PXXXX, type(args.PXXXX))
print(args.OXXXX, type(args.OXXXX))
GXXXX = args.GXXXX
PXXXX = args.PXXXX
OXXXX = args.OXXXX

print(args.GXXXXPLUSONE, type(args.GXXXXPLUSONE))
GXXXXPLUSONE = args.GXXXXPLUSONE




def random_sampling_mols(All_mols, sample_number=1000):
    #------random sampling 1000 mols------------------------------------------
    print("The length of All_mols is : ", len(All_mols))
    if (sample_number >= 1) & ( len(All_mols) >= sample_number):
        sample_indexs = np.random.choice(len(All_mols), sample_number, replace=False)
        All_mols_sample = [All_mols[idx] for idx in sample_indexs]
        All_mols = All_mols_sample 
        print("After sampling, the length of All_mols is : ", len(All_mols))
    else:
        print("The length of All_mols is  smaller than sample_number: ", len(All_mols) , "   <   ",  sample_number )
    
    return All_mols


#------ main function -----------------------------------------------------
#--------------------------------------------------------------------------

#----- reading opt_mols sdf file -----------------
opt_sdf_path = './project/'  +  GXXXX  +   '/'  +  OXXXX   + '.sdf'
sdsuppl = Chem.SDMolSupplier(opt_sdf_path)
opt_mols = [x for x in sdsuppl if x is not None]
print("length of opt_mols is : ", len(opt_mols))


#-----------------------------------------------------------------------------
#---- create combinational compound library --------------------------
#---- via calling mutations_N_to_CH_aromatic() or debranch_CR_aromatic()

print("\n", '*'*80)

All_mols = []
for opt_mol  in opt_mols[:10]:
    # mutation_mols = mutations_CH_aromatic(opt_mol, 
    #                                       patt=Chem.MolFromSmarts('c[H]'), 
    #                                       substitute_atom=CARBON_SUB,
    #                                       substitute_atomnum=7, 
    #                                       substi_number=1, 
    #                                       sample_number=3, 
    #                                       #is_showmol=True, 
    #                                       max_subpos=10
    #                                       )

    mutation_mols = mutations_N_to_CH_aromatic(opt_mol, 
                                                patt=Chem.MolFromSmarts( N_PATT ), 
                                                #patt=Chem.MolFromSmarts('[n&R&H0,N&R&H0]'),  
                                                #patt=Chem.MolFromSmarts('[n&R&H0]'),  
                                                #patt=Chem.MolFromSmarts('[n&R]'),  
                                                #patt=Chem.MolFromSmarts('[n&R&H0,N&R]'), 
                                                substitute_group=Chem.MolFromSmarts('C-[H]'),
                                                substi_site_symbol='C',
                                                substi_number=1, 
                                                sample_number=2, 
                                                is_showmol=True,
                                                #is_showmol=False,
                                                max_subpos=10
                                                )

    
    #All_mols +=  [ opt_mol ] +  mutation_mols 
    
    # for termino_hydrogen_sub in TERMINO_HYDROGEN_SUBS:
    #     mutation_mols1 = mutations_CH_aromatic_termino_H(opt_mol, 
    #                                                     patt=Chem.MolFromSmarts('c[H]'), 
    #                                                     substitute_group=Chem.MolFromSmiles( termino_hydrogen_sub ), 
    #                                                     substi_number=1, 
    #                                                     sample_number=3, 
    #                                                     #is_showmol=True,
    #                                                     max_subpos=10
    #                                                     )


    mutation_mols1 = debranch_CR_aromatic(opt_mol, 
                                          patt=Chem.MolFromSmarts( CR_PATT ),      
                                          #patt=Chem.MolFromSmarts('[a&R]-[A&!#1,a&!#1]'),  
                                          dummylabels=(11111, 22222),
                                          #dummylabels=(0, 0),
                                          substi_atom="H", 
                                          substi_atomnum=1, 
                                          # substi_atom="F", 
                                          # substi_atomnum=9,
                                          debranch_number=2, 
                                          sample_number=8, 
                                          is_showmol=True,
                                          #is_showmol=False,
                                          max_debranchpos=10
                                        )


        
    All_mols +=  [ opt_mol ] +  mutation_mols + mutation_mols1
    

print("The length of All_mols is : ", len(All_mols))


draw_mols(All_mols, 
          16, 
          './project/%s/images/foo011a--mutation-mols-demo.png'%GXXXX, 
          molsPerRow=4, 
          size=(800, 800)
          )


All_mols  =  delete_repeated_smiles_of_mols(All_mols)
print("After delete_repeated_smiles_of_mols, the length of All_mols is : ", len(All_mols))

#------random sampling 1000 mols-----------------------------------------------------
All_mols = random_sampling_mols(All_mols, sample_number=1000)

cpd_names = [ "cpd-%05d" % (i+1) for i in range(len(All_mols)) ]
for mol, cpd_name in zip(All_mols, cpd_names):
    mol.SetProp("_Name", cpd_name )


#---- write molecules to SD File ------------------------------- 
call_mkdir_GXXXXplus1 = 'mkdir  ./project/' +  GXXXXPLUSONE  +  '/  2>/dev/null  '
status, output = subprocess.getstatusoutput( call_mkdir_GXXXXplus1 )
print("mkdir  ./project/GXXXX+1/;    status, output = ", status, output)

outfile_path = './project/'   +   GXXXXPLUSONE   +  '/'   + GXXXXPLUSONE + 'a.sdf'
All_mols_w = Chem.SDWriter(outfile_path)
for m in tqdm(All_mols): 
    m_temp = Chem.AddHs(m)
    m1_temp, nb_try_EmbedMolecule = try_embedmolecule(m_temp)
    All_mols_w.write(m1_temp)




print("\n### End of program: PG012a!\n")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 16 14:21:51 2021

This is the global drive script of the so-called 
SALAM algorithm (Self-consistent generation Algorithm for vitual compound Library of high-Abundance target Materials)
for machine learning and mutation driven molecular property optimization.


@author: tucy
"""


# import sys 
# import os
# main_wkdir = os.getcwd()
# sys.path.append(main_wkdir)


import os
import time
import subprocess
import datetime

from salam.module_MD.Metrics import percentage_lt_criticval
from salam.module_MD.Metrics import accumulate_optmols 
from salam.module_MD.Metrics import accumulate_optmols_S1En
from salam.module_MD.Metrics import accumulate_optmols_intersection
# from salam.module_MD.Metrics import get_allfrags_via_bricsdecompose
from salam.module_MD.Metrics import get_num_replaceable_aromatic_CHs
from salam.module_MD.Metrics import get_paras_dic
from salam.module_MD.Metrics import convert_str_to_list
from salam.module_MD.Metrics import sort_mols_by_SAScores
from salam.module_MD.Metrics import plot_muation_diagram


import pandas as pd
from matplotlib import pyplot as plt

import argparse

parser = argparse.ArgumentParser(description="manual to this script: ")

parser.add_argument('--inp', 
                    type=str, 
                    default='./salam_paras.inp', 
                    required=False, 
                    metavar='--input_filename', 
                    help='The input filename which stores the parameters for running of the program.')

parser.add_argument('--output', 
                    type=str, 
                    default='./SALAM.log', 
                    required=False, 
                    metavar='--output_filename', 
                    help='The output filename which stores the result of the program.')

parser.add_argument('--restart_maxminpicker', type=bool, default=False )
parser.add_argument('--restart_generation', type=int, default=-1)
parser.add_argument('--restart_optjob', type=bool, default=False )
parser.add_argument('--restart_propjob', type=bool, default=False )


args = parser.parse_args()

print("input: ", args.inp, " ; the type is ", type(args.inp) )
INPUT_FILE = args.inp

print("output: ", args.output, " ; the type is ", type(args.output) )
OUTPUT_FILE = args.output

print("restart_maxminpicker: ", args.restart_maxminpicker, " ; the type is ", type(args.restart_maxminpicker) )
RESTART_MAXMINPICKER = bool(args.restart_maxminpicker)

print("restart_generation: ", args.restart_generation, " ; the type is ", type(args.restart_generation) )
RESTART_GENERATION = args.restart_generation

print("restart_optjob: ", args.restart_optjob, " ; the type is ", type(args.restart_optjob) )
RESTART_OPTJOB = bool(args.restart_optjob)

print("restart_propjob: ", args.restart_propjob, " ; the type is ", type(args.restart_propjob) )
RESTART_PROPJOB = bool(args.restart_propjob)


#------- get parameters via calling get_paras_dic -----------------------------
PARAS_DIC = get_paras_dic(filename=INPUT_FILE)

SLEEP_SECONDS = int( PARAS_DIC['SLEEP_SECONDS'] )

# GXXXX = 'G%04d'%PARAS_DIC['MUTATION_GENERATION']
# PXXXX = 'P%04d'%PARAS_DIC['MUTATION_GENERATION']
# OXXXX = 'O%04d'%PARAS_DIC['MUTATION_GENERATION']
# GXXXXPLUSONE = 'G%04d'%(PARAS_DIC['MUTATION_GENERATION'] +1)


#-----------------------------------------------------------------------------
SALAM_LOG = OUTPUT_FILE                          #  "./SALAM.log"  # the output log for SALAM.py 
module_name = 'moduleMD'
src_dir = 'src/'

TERMINO_HYDROGEN_SUBS =  convert_str_to_list(str1=PARAS_DIC['TERMINO_HYDROGEN_SUBS']) 

TERMINO_HYDROGEN_SUBS_CMD_PARA = "  "
for i in range(len(TERMINO_HYDROGEN_SUBS)): 
    TERMINO_HYDROGEN_SUBS_CMD_PARA += "  --termino_hydrogen_subs='%s' "% TERMINO_HYDROGEN_SUBS[i]


MAX_SUBPOS              =  int( PARAS_DIC['MAX_SUBPOS'] )
SUBSTI_NUMBER_CARBON    =  int( PARAS_DIC['SUBSTI_NUMBER_CARBON'] )
SAMPLE_NUMBER_CARBON    =  int( PARAS_DIC['SAMPLE_NUMBER_CARBON'] )
SUBSTI_NUMBER_HYDROGEN  =  int( PARAS_DIC['SUBSTI_NUMBER_HYDROGEN'] )
SAMPLE_NUMBER_HYDROGEN  =  int( PARAS_DIC['SAMPLE_NUMBER_HYDROGEN'] )

mutation_ctl_para_names = [ 'max_subpos', 'substi_number_carbon', 'sample_number_carbon', 'substi_number_hydrogen', 'sample_number_hydrogen' ]
mutation_ctl_para_values = [ MAX_SUBPOS, SUBSTI_NUMBER_CARBON, SAMPLE_NUMBER_CARBON, SUBSTI_NUMBER_HYDROGEN, SAMPLE_NUMBER_HYDROGEN ]

MUTATION_CONTROL_CMD_PARA = "  "
for name1, value1 in zip(mutation_ctl_para_names, mutation_ctl_para_values): 
    MUTATION_CONTROL_CMD_PARA += "   --%s=%d   "%(name1, value1)


#-----------------------------------------------------------------------------
PROG_LIST0 = [
            'PG001__Generate_G0000.py',
            'PG002__MaxMinPicker.py',
            'PG003__bat_sdf_to_gscom.py',
            'PG004__bat_qsub_jobs_opt.py',
            'PG005__optimization_analysis_completeness.py',
            'PG006__bat_g09log_to_propcom.py',
            'PG007__bat_qsub_jobs_prop.py',
            'PG008__property_analysis_completeness.py',
            'PG009__property_analysis_summary.py',
            'PG010__featurizers_train_evaluate.py',
            'PG011__predict_opt_molecules.py',
            'PG012__Generate_Gnew.py',    
            # 'PG012a__Generate_Gnew_Debranch.py',
            'PG013__MurckoScaffold.py',
            'PG014__Energy_Sieve.py',
            'PG015__Map-mols-by-FPSimilarity.py'
            ]

PROG_LIST = [ './' + src_dir + x  for x in PROG_LIST0]


#--------- definitions for functions ------------------------------------------
#------------------------------------------------------------------------------
def call_cmd_via_subprocess_run(cmd,
                                loop_time_limit=10000,
                                sleep_seconds=SLEEP_SECONDS
                                ):
    loop_time = 0
    returncode = -9999
    print("calling call_cmd_via_subprocess_run.")
    while( (returncode != 0)  and (loop_time < loop_time_limit) ):
        try:
            print("loop_time = ", loop_time)
            cmd_CompletedProcess = subprocess.run(args=cmd,
                                                  stdout=subprocess.PIPE,
                                                  stderr=subprocess.PIPE,
                                                  shell=True,
                                                  check=True
                                                  )
            #cmd = cmd_CompletedProcess.args
            returncode = cmd_CompletedProcess.returncode
            stdoutput = cmd_CompletedProcess.stdout.decode()
            
            print("Normal call.")
            #print("cmd = ", cmd)
            print("in correct branch: loop_time = ", loop_time, " , returncode = ", returncode)
            print("stdout is: ")
            print(stdoutput)

            
        except subprocess.CalledProcessError as e:
            returncode = e.returncode
            stderr = e.stderr.decode()
            print("subprocess.CalledProcessError")
            print("stderr : ", stderr)
        
            print("in error branch:  loop_time = ", loop_time, " , returncode = ", returncode)
            print("sleep for a while, and retry.")
            time.sleep(sleep_seconds)
            
            #loop_time += 1
            
        # except subprocess.TimeoutExpired as t:
        #     returncode = t.returncode
        #     stderr = t.stderr.decode()
            
        #     print("subprocess.TimeoutExpired")
        #     print("stderr : ", stderr)
        
        loop_time += 1
        print("outside try:  loop_time = ", loop_time, " , returncode = ", returncode)
        

def run_PG001():
    #------- run PG001 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[0],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[0]  +   "  '      1>  "   +  SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para1 =  "python   " +  PROG_LIST[0]   +  "      1>>  "    +   SALAM_LOG
    os.system( call_prog_with_para1 )
    #------- end run PG001  PG001__Generate_G0000.py------------------------------------------------------------------


def run_PG002():
    #------- run PG002 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[1],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[1]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para2 =  "python   " +  PROG_LIST[1]  \
        +   "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    +  "    --pickmols_size=%d"% int( PARAS_DIC['PICKMOLS_SIZE'] ) \
                        +  "      1>>    "    +   SALAM_LOG
    
    os.system( call_prog_with_para2 )
    #------- end run PG002  PG002__MaxMinPicker.py------------------------------------------------------------------
    

def run_PG003():
    #------- run PG003 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[2],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[2]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para3 =  "python   " +  PROG_LIST[2]  \
        +   "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    +  "   --gs_opt_method=%s"%PARAS_DIC['GS_OPT_METHOD']  \
                        + "  --remote_wkdir=%s"%PARAS_DIC['REMOTE_WKDIR']   \
                            + "   --num_serviers=%d"% int( PARAS_DIC['NUM_SERVIERS'] )  \
                            +   "      1>>    "    +   SALAM_LOG
                        
    os.system( call_prog_with_para3 )
    #------- end run PG003  PG003__bat_sdf_to_gscom.py------------------------------------------------------------------





def update_acc_computedmols(acc_computedmols='./project/G0002/acc_computedmols_opt.csv',
                            old_acc_computedmols='./project/G0001/acc_computedmols_opt.csv',
                            new_computedmols='./project/G0001/P0001rev.sdf'):
    
    return 0
    
    


def cpdnames_checker(acc_computedmols='./project/G0001/acc_computedmols_opt.csv',
                    computingmols='./project/G0001/computingmols_opt.csv'):
    print("### Begin cpdnames_checker. Check whether computing_mols were in acc_computed_mols.")
    
    #  acc_computedmols_opt.csv 应该是这样一个逗号分割值文件， 包括两列， smiles 和 filepath 仅仅用首次（最原始的）被计算过的
    #  文件路径（字符串, 不包括文件名后缀）。  例如， 
    #  c1ccc(O)cc1， ‘./project/G0000/pick_mols/com-files/cpd-00016’
    #  
    #  类似地，computingmols_opt.csv 应该由 对应的 P0001.sdf 生成， 其结构为，
    #  c1ccc(O)cc1， ‘./project/G0001/pick_mols/com-files/cpd-00098’
    #
    #  计算两者的交集，获得对应的已计算过的cpdnames含路径，  将交集中的文件 (log) 拷贝到新的要计算的 opt_jobs 工作目录下，
    #  使用系统拷贝命令， cp   filepath1   to  filepath2.   
    #  
    #  主要是第一个csv文件，需要持续更新， 以便不漏过新增的分子。   
    
    
    print("### End of cpdnames_checker")
    
    return 0
    

    

def run_PG004():
    #------- run PG004 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[3],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[3]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para4 =  "python   " +  PROG_LIST[3]  \
        +   "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    + "  --remote_wkdir=%s"%PARAS_DIC['REMOTE_WKDIR']   \
                        + "   --num_serviers=%d"% int( PARAS_DIC['NUM_SERVIERS'] )  \
                        + "      1>>    "    +   SALAM_LOG
                    
    os.system( call_prog_with_para4 )
    #------- end run PG004  PG004__bat_qsub_jobs_opt.py------------------------------------------------------------------    


def run_PG005(call_manner=1):
    #------- run PG005 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[4],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[4]  +   "  '      1>>    "    +   SALAM_LOG
    retcode = os.system( call_echo_run_prog_to_salam_log )
    print("retcode = ", retcode)
    
    call_prog_with_para5 =  "python   " +  PROG_LIST[4]  \
                            +   "   --GXXXX=%s"%GXXXX  \
                                +  "   --PXXXX=%s"%PXXXX  \
                                    +  "   --OXXXX=%s"%OXXXX   \
                                        + "  --remote_wkdir=%s"%PARAS_DIC['REMOTE_WKDIR']   \
                                            + "   --num_serviers=%d"% int( PARAS_DIC['NUM_SERVIERS'] )  \
                                            +  "   --pbs_job_incompleteness_tolerance=%f"% float( PARAS_DIC['PBS_JOB_INCOMPLETENESS_TOLERANCE_OPT'] )  \
                                              + "   --sleep_seconds=%d"% int( PARAS_DIC['OPT_SLEEP_SECONDS'] )  \
                                                  + "      1>>    "    \
                                                      +   SALAM_LOG
    
    if (call_manner == 1): 
        call_cmd_via_subprocess_run(cmd=call_prog_with_para5,
                                    loop_time_limit=10000,
                                    sleep_seconds=SLEEP_SECONDS
                                    )              
    elif (call_manner == 2):
        retcode =  os.system( call_prog_with_para5 )
        print("retcode = ", retcode)
        
    else:
        print("call_manner must be 1 or 2. 1 for subprocess.run, 2 for os.system")
        
    #------- end run PG005 PG005__optimization_analysis_completeness.py------------------------------------------------------------------
    

def run_PG006():    
    #------- run PG006 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[5],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[5]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para6 =  "python   " +  PROG_LIST[5]  \
        +   "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    +  "   --calc_prop=%s"%PARAS_DIC['CALC_PROP']    \
                        + "   --prop_calc_method=%s"%PARAS_DIC['PROP_CALC_METHOD']  \
                            + "  --remote_wkdir=%s"%PARAS_DIC['REMOTE_WKDIR']   \
                                + "   --num_serviers=%d"% int( PARAS_DIC['NUM_SERVIERS'] )  \
                                + "      1>>    "    +   SALAM_LOG
                        
    os.system( call_prog_with_para6 )
    #------- end run PG006 PG006__bat_g09log_to_propcom.py------------------------------------------------------------------
    

def run_PG007():
    #------- run PG007 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[6],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[6]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para7 =  "python   " +  PROG_LIST[6]  \
        +   "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    + "  --remote_wkdir=%s"%PARAS_DIC['REMOTE_WKDIR']   \
                        + "   --num_serviers=%d"% int( PARAS_DIC['NUM_SERVIERS'] )  \
                        + "      1>>    "    +   SALAM_LOG
                    
    os.system( call_prog_with_para7 )
    #------- end run PG007 PG007__bat_qsub_jobs_prop.py------------------------------------------------------------------
    

def run_PG008(call_manner=1):
    #------- run PG008 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[7],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[7]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para8 =  "python   " +  PROG_LIST[7]  \
        +  "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    +  "   --pbs_job_incompleteness_tolerance=%f"% float( PARAS_DIC['PBS_JOB_INCOMPLETENESS_TOLERANCE_PROP'] ) \
                        +  "   --sleep_seconds=%d"% int( PARAS_DIC['PROP_SLEEP_SECONDS'] ) \
                            + "  --remote_wkdir=%s"%PARAS_DIC['REMOTE_WKDIR']   \
                                + "   --num_serviers=%d"% int( PARAS_DIC['NUM_SERVIERS'] )  \
                                + "      1>>    "    +   SALAM_LOG
                                
    if (call_manner == 1): 
        call_cmd_via_subprocess_run(cmd=call_prog_with_para8,
                                    loop_time_limit=100000,
                                    sleep_seconds=SLEEP_SECONDS
                                    )              
    elif (call_manner == 2):
        os.system( call_prog_with_para8 )
    else:
        print("call_manner must be 1 or 2. 1 for subprocess.run, 2 for os.system")
    #------- end run PG008  PG008__property_analysis_completeness.py------------------------------------------------------------------   


def run_PG009(call_manner=1):
    #------- run PG009 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[8],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[8]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para9 =  "python   " +  PROG_LIST[8]  \
        +  "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    + "   --analyzed_property=%s"%PARAS_DIC['ANALYZED_PROPERTY']  \
                        +  "   --pbs_job_incompleteness_tolerance=%f"% float( PARAS_DIC['PBS_JOB_INCOMPLETENESS_TOLERANCE_PROP'] ) \
                            +  "   --sleep_seconds=%d"%SLEEP_SECONDS  \
                                + "      1>>    "    +   SALAM_LOG
                                
    if (call_manner == 1): 
        call_cmd_via_subprocess_run(cmd=call_prog_with_para9,
                                    loop_time_limit=100000,
                                    sleep_seconds=SLEEP_SECONDS
                                    )              
    elif (call_manner == 2):
        os.system( call_prog_with_para9 )
    else:
        print("call_manner must be 1 or 2. 1 for subprocess.run, 2 for os.system")
    #------- end run PG009  PG009__property_analysis_summary.py------------------------------------------------------------------



def run_PG010(call_manner=1):
    #------- run PG010 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[9],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[9]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para10 =  "python   " +  PROG_LIST[9]  \
        +   "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    + "   --analyzed_property=%s"%PARAS_DIC['ANALYZED_PROPERTY']  \
                        + "      1>>    "    +   SALAM_LOG
                    
    if (call_manner == 1): 
        call_cmd_via_subprocess_run(cmd=call_prog_with_para10,
                                    loop_time_limit=10000,
                                    sleep_seconds=SLEEP_SECONDS
                                    )              
    elif (call_manner == 2):
        os.system( call_prog_with_para10 )
    else:
        print("call_manner must be 1 or 2. 1 for subprocess.run, 2 for os.system")
    #------- end run PG010  PG010__featurizers_train_evaluate.py------------------------------------------------------------------ 
    

def run_PG011(call_manner=1):    
    #------- run PG011 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[10],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[10]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para11 =  "python   " +  PROG_LIST[10]  \
        +   "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    +  "   --prop_tendency=%s"%PARAS_DIC['PROP_TENDENCY']  \
                        + "      1>>    "    +   SALAM_LOG
                    
    if (call_manner == 1): 
        call_cmd_via_subprocess_run(cmd=call_prog_with_para11,
                                    loop_time_limit=10000,
                                    sleep_seconds=SLEEP_SECONDS
                                    )              
    elif (call_manner == 2):
        os.system( call_prog_with_para11 )
    else:
        print("call_manner must be 1 or 2. 1 for subprocess.run, 2 for os.system")
    #------- end run PG011  PG011__predict_opt_molecules.py------------------------------------------------------------------
    

def run_PG012(call_manner=1):
    #------- run PG012 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[11],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[11]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )

    call_prog_with_para12 =  "python   " +  PROG_LIST[11]  \
        +   "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    +  "   --GXXXXPLUSONE=%s"%GXXXXPLUSONE  \
                        + "   --carbon_sub=%s"%PARAS_DIC['CARBON_SUB']   \
                            + TERMINO_HYDROGEN_SUBS_CMD_PARA \
                                + MUTATION_CONTROL_CMD_PARA \
                                    + "  --is_substi_carbon=%d  "% int( PARAS_DIC['IS_SUBSTI_CARBON'] )   \
                                        + "  --is_substi_hydrogen=%d  "% int( PARAS_DIC['IS_SUBSTI_HYDROGEN'] )   \
                                            + "  --parent_optmols_size=%d   "% int( PARAS_DIC['PARENT_OPTMOLS_SIZE'] )  \
                                                + "      1>>    "    +   SALAM_LOG
                            
                            
    if (call_manner == 1): 
        call_cmd_via_subprocess_run(cmd=call_prog_with_para12,
                                    loop_time_limit=10000,
                                    sleep_seconds=SLEEP_SECONDS
                                    )              
    elif (call_manner == 2):
        os.system( call_prog_with_para12 )
    else:
        print("call_manner must be 1 or 2. 1 for subprocess.run, 2 for os.system")
    #------- end run PG012 PG012__Generate_Gnew.py------------------------------------------------------------------


def run_PG012a(call_manner=1):
    #------- run PG012a with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[12],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[12]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )

    call_prog_with_para12a =  "python   " +  PROG_LIST[12]  \
        +   "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    +  "   --GXXXXPLUSONE=%s"%GXXXXPLUSONE \
                        + "  --N_PATT='%s' "%(PARAS_DIC['N_PATT'])  \
                            + "  --CR_PATT='%s' "%(PARAS_DIC['CR_PATT'])  \
                                + "      1>>    "    +   SALAM_LOG
                        # +  "--N_PATT=%s"%N_PATT \
                        #     +  "--CR_PATT=%s"%CR_PATT \
                            
                            
    if (call_manner == 1): 
        call_cmd_via_subprocess_run(cmd=call_prog_with_para12a,
                                    loop_time_limit=10000,
                                    sleep_seconds=SLEEP_SECONDS
                                    )              
    elif (call_manner == 2):
        os.system( call_prog_with_para12a )
    else:
        print("call_manner must be 1 or 2. 1 for subprocess.run, 2 for os.system")
    #------- end run PG012a  PG012a__Generate_Gnew_Debranch.py------------------------------------------------------------------



def run_PG013(call_manner=1):
    #------- run PG013 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[12],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[12]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para13 =  "python   " +  PROG_LIST[12]  \
        +   "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    +  "   --num_high_freqs=%d   "% int(PARAS_DIC['NUM_HIGH_FREQS'])   \
                        + "      1>>    "    +   SALAM_LOG
                    
    if (call_manner == 1): 
        call_cmd_via_subprocess_run(cmd=call_prog_with_para13,
                                    loop_time_limit=10000,
                                    sleep_seconds=SLEEP_SECONDS
                                    )              
    elif (call_manner == 2):
        os.system( call_prog_with_para13 )
    else:
        print("call_manner must be 1 or 2. 1 for subprocess.run, 2 for os.system")
    #------- end run PG013  PG013__MurckoScaffold.py------------------------------------------------------------------ 
    


def run_PG014(call_manner=1):
    #------- run PG014 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[13],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[13]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para14 =  "python   " +  PROG_LIST[13]  \
        +   "   --GXXXX=%s"%GXXXX  \
            +  "   --PXXXX=%s"%PXXXX  \
                +  "   --OXXXX=%s"%OXXXX   \
                    +  "   --energy_tendency=%s  "%  PARAS_DIC['ENERGY_TENDENCY']   \
                        +  "   --energy_threshold1=%f   "% float(PARAS_DIC['ENERGY_THRESHOLD1'])   \
                            +  "   --energy_threshold2=%f   "% float(PARAS_DIC['ENERGY_THRESHOLD2'])   \
                                +  "   --stokes_shift=%f   "% float(PARAS_DIC['STOKES_SHIFT'])   \
                                    + "      1>>    "    +   SALAM_LOG
                    
    if (call_manner == 1): 
        call_cmd_via_subprocess_run(cmd=call_prog_with_para14,
                                    loop_time_limit=10000,
                                    sleep_seconds=SLEEP_SECONDS
                                    )              
    elif (call_manner == 2):
        os.system( call_prog_with_para14 )
    else:
        print("call_manner must be 1 or 2. 1 for subprocess.run, 2 for os.system")
    #------- end run PG014  PG014__Energy_Sieve.py------------------------------------------------------------------ 


def run_PG015(call_manner=1):
    #------- run PG015 with paras -----------------------------------------------------------
    print("### Run  %s   1>    %s \n"%(PROG_LIST[14],  SALAM_LOG) )
    
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   PROG_LIST[14]  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    call_prog_with_para15 =  "python   " +  PROG_LIST[14]  \
                            +   "   --GXXXX=%s"%GXXXX  \
                                + "      1>>    "    +   SALAM_LOG
                    
    if (call_manner == 1): 
        call_cmd_via_subprocess_run(cmd=call_prog_with_para15,
                                    loop_time_limit=10000,
                                    sleep_seconds=SLEEP_SECONDS
                                    )              
    elif (call_manner == 2):
        os.system( call_prog_with_para15 )
    else:
        print("call_manner must be 1 or 2. 1 for subprocess.run, 2 for os.system")
    #------- end run PG015  PG015__Map-mols-by-FPSimilarity.py------------------------------------------------------------------ 



def run_evaluate_abundance(call_manner=1):
        print("### Run  run_evaluate_abundance   1>>    %s \n" %SALAM_LOG )
        call_echo_run_prog_to_salam_log = "echo  '### run "  +   "run_evaluate_abundance"  +   "  '      1>>    "    +   SALAM_LOG
        os.system( call_echo_run_prog_to_salam_log )
        
        GXXXX_df_ECFP_csv_path = './project/'  +  GXXXX    +  '/'  + GXXXX   +  '_df_ECFP.csv'

        call_prog_with_para_eva_abu = 'test -s  '   +   GXXXX_df_ECFP_csv_path  +  '   ' 
        
        #print("\nTest if file %s exist?\n" % GXXXX_df_ECFP_csv_path)
        print("\nTest if file %s exist and not empty?\n" % GXXXX_df_ECFP_csv_path)
        if (call_manner == 1): 
            call_cmd_via_subprocess_run(cmd=call_prog_with_para_eva_abu,
                                        loop_time_limit=1000,
                                        sleep_seconds=SLEEP_SECONDS
                                        ) 
        elif (call_manner == 2):
            os.system( call_prog_with_para_eva_abu )
        else:
            print("call_manner must be 1 or 2. 1 for subprocess.run, 2 for os.system")
        
        abundance = percentage_lt_criticval(filename=GXXXX_df_ECFP_csv_path, 
                                            critical_value=0.15
                                            )
        
        abundances.append(abundance)
        
        with open(SALAM_LOG, 'a+') as salam_log:
            print("\nGeneration = %4d , abundance = %10.4f\n" %(MUTATION_GENERATION, abundance) )
            salam_log.write("\n Generation = %4d , abundance = %10.4f\n" %(MUTATION_GENERATION, abundance) )
            
            if (len(abundances) >= 2):
                print("The varation on abundances is: %10.4f\n"%(abundances[-1]  -  abundances[-2] ) )
                salam_log.write("The varation on abundances is: %10.4f\n"%(abundances[-1]  -  abundances[-2] ) )
            else:
                print('')
            
            salam_log.write("\n### End of run_evaluate_abundance.\n")
            

def run_accumulate_optmols( ):
    print("### Run  run_accumulate_optmols   1>>    %s \n" %SALAM_LOG )
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   "run_accumulate_optmols"  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
        
    num_acc_optmol = accumulate_optmols(mutation_generation=MUTATION_GENERATION, 
                                        old_acc_optmols_smi_path='./project/G%04d/acc_optmols.smi'%( MUTATION_GENERATION -1 ),
                                        new_mols_path='./project/G%04d/G%04d_df_ECFP.csv'%( MUTATION_GENERATION , MUTATION_GENERATION ),
                                        acc_optmols_smi_path='./project/G%04d/acc_optmols.smi'%( MUTATION_GENERATION ), 
                                        column_names=[ 'smiles', 'prop_preds' ], 
                                        critical_value=0.15,
                                        is_write_sdf=False, 
                                        outdir_path = './project/G%04d/acc_optmols/'%( MUTATION_GENERATION )
                                        )
    
    num_acc_optmols.append(num_acc_optmol)

    with open(SALAM_LOG, 'a+') as salam_log:
        print("\nGeneration = %4d , num_acc_optmol = %10d\n" %(MUTATION_GENERATION, num_acc_optmol) )
        salam_log.write("\n Generation = %4d , num_acc_optmol = %10d\n" %(MUTATION_GENERATION, num_acc_optmol) )
        
        if (len(num_acc_optmols) >= 2):
            print("The varation on num_acc_optmol is: %10d\n"%(num_acc_optmols[-1]  -  num_acc_optmols[-2] ) )
            salam_log.write("The varation on num_acc_optmol is: %10d\n"%(num_acc_optmols[-1]  -  num_acc_optmols[-2] ) )
        else:
            print('')
        
        salam_log.write("\n### End of run_accumulate_optmols.\n")    



def run_sort_mols_by_SAScores():
    #----- call sort_mols_by_SAScores ---------------------------------------------
    print("### Run  sort_mols_by_SAScores   1>>    %s \n" %SALAM_LOG )
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   "sort_mols_by_SAScores"  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
    
    print("MUTATION_GENERATION_LIMIT = ", int( PARAS_DIC['MUTATION_GENERATION_LIMIT'] ) )
    print("After minus 1, the current value of MUTATION_GENERATION = ", MUTATION_GENERATION )

    # os.path.exists, os.path.isfile
    if os.path.isfile('./project/G%04d/acc_optmols.smi'%(MUTATION_GENERATION) ):
        sort_mols_by_SAScores(mutation_generation= (MUTATION_GENERATION),
                              acc_optmols_smi_path='./project/G%04d/acc_optmols.smi'%(MUTATION_GENERATION), 
                              column_names=[ 'mols', 'smiles' ], 
                              outdir_path='./project/G%04d/acc_optmols_sortedbysas/'%(MUTATION_GENERATION),
                              is_write_sdf=True, 
                              new_sdf_path='./project/G%04d/acc_optmols_sortedbysas.sdf'%(MUTATION_GENERATION),
                              ascending=True
                              )
    else:
        print("\n!!! required acc_optmols.smi does not exist. Skip transformation.\n")
    
    print("### End of sort_mols_by_SAScores   1>>    %s \n" %SALAM_LOG )
    call_echo_run_prog_to_salam_log = "echo  '### End of "  +   "sort_mols_by_SAScores"  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )



def run_accumulate_optmols_S1En( ):
    print("### Run  run_accumulate_optmols_S1En   1>>    %s \n" %SALAM_LOG )
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   "run_accumulate_optmols_S1En"  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )

    num_acc_optmol_S1En_R, num_acc_optmol_S1En_G, num_acc_optmol_S1En_B = accumulate_optmols_S1En(mutation_generation=MUTATION_GENERATION, 
                                                                        old_acc_optmols_smi_paths=[ './project/G%04d/acc_optmols--S1En-%s.smi'%( MUTATION_GENERATION -1, color ) for color in ["R", "G", "B"] ],
                                                                        new_mols_path='./project/G%04d/G%04d_df_ECFP--S1En.csv'%( MUTATION_GENERATION, MUTATION_GENERATION),
                                                                        acc_optmols_smi_paths=[ './project/G%04d/acc_optmols--S1En-%s.smi'%( MUTATION_GENERATION, color ) for color in ["R", "G", "B"] ],
                                                                        column_names=[ 'compoundnames', 'smiles', 'prop_preds' ], 
                                                                        critical_values=[ float(PARAS_DIC['ENERGY_THRESHOLD1']) + float(PARAS_DIC['STOKES_SHIFT']),
                                                                                         float(PARAS_DIC['ENERGY_THRESHOLD2']) + float(PARAS_DIC['STOKES_SHIFT']) ],
                                                                        )

    num_acc_optmol_S1En_RGB = [num_acc_optmol_S1En_R, num_acc_optmol_S1En_G, num_acc_optmol_S1En_B]
    
    num_acc_optmols_S1En.append(num_acc_optmol_S1En_RGB)

    with open(SALAM_LOG, 'a+') as salam_log:
        print("\nGeneration = %4d , num_acc_optmol_S1En_RGB = "%(MUTATION_GENERATION), num_acc_optmol_S1En_RGB )
        salam_log.write("\n Generation = %4d , num_acc_optmol_S1En_RGB = %5d %5d %5d" \
                        %(MUTATION_GENERATION, num_acc_optmol_S1En_R, num_acc_optmol_S1En_G, num_acc_optmol_S1En_B) ) 
        
        if (len(num_acc_optmols_S1En) >= 2):
            print("\nThe varation on num_acc_optmol_S1En_R is: %10d\n"%(num_acc_optmols_S1En[-1][0]  -  num_acc_optmols_S1En[-2][0] ) )
            print("\nThe varation on num_acc_optmol_S1En_G is: %10d\n"%(num_acc_optmols_S1En[-1][1]  -  num_acc_optmols_S1En[-2][1] ) )
            print("\nThe varation on num_acc_optmol_S1En_B is: %10d\n"%(num_acc_optmols_S1En[-1][2]  -  num_acc_optmols_S1En[-2][2] ) )
            salam_log.write("\nThe varation on num_acc_optmol_S1En is: %10d\n"%(num_acc_optmols_S1En[-1][0]  -  num_acc_optmols_S1En[-2][0] ) )
            salam_log.write("\nThe varation on num_acc_optmol_S1En is: %10d\n"%(num_acc_optmols_S1En[-1][1]  -  num_acc_optmols_S1En[-2][1] ) )
            salam_log.write("\nThe varation on num_acc_optmol_S1En is: %10d\n"%(num_acc_optmols_S1En[-1][2]  -  num_acc_optmols_S1En[-2][2] ) )
        else:
            print('')
        
        salam_log.write("\n### End of run_accumulate_optmols_S1En.\n")  



def run_accumulate_optmols_intersection( ):
    print("### Run  run_accumulate_optmols_intersection   1>>    %s \n" %SALAM_LOG )
    call_echo_run_prog_to_salam_log = "echo  '### run "  +   "run_accumulate_optmols_intersection"  +   "  '      1>>    "    +   SALAM_LOG
    os.system( call_echo_run_prog_to_salam_log )
        
    num_acc_optmol_INTERSECTION = accumulate_optmols_intersection(mutation_generation=MUTATION_GENERATION, 
                                                                    acc_optmols_smi_path1='./project/G%04d/acc_optmols.smi'%( MUTATION_GENERATION ),
                                                                    acc_optmols_smi_path2='./project/G%04d/acc_optmols--S1En-B.smi'%( MUTATION_GENERATION ),
                                                                    acc_optmols_smi_path3='./project/G%04d/acc_optmols--INTERSCTION-B.smi'%( MUTATION_GENERATION ), 
                                                                    is_write_sdf=False, 
                                                                    outdir_path='./project/G%04d/acc_optmols--INTERSCTION/'%( MUTATION_GENERATION )
                                                                    )
    
    num_acc_optmols_INTERSECTIONs.append(num_acc_optmol_INTERSECTION)

    with open(SALAM_LOG, 'a+') as salam_log:
        print("\nGeneration = %4d , num_acc_optmol_INTERSECTION = %10d\n" %(MUTATION_GENERATION, num_acc_optmol_INTERSECTION ) )
        salam_log.write("\n Generation = %4d , num_acc_optmol_INTERSECTION = %10d\n" %(MUTATION_GENERATION, num_acc_optmol_INTERSECTION ) )
        
        if (len(num_acc_optmols_S1En) >= 2):
            print("The varation on num_acc_optmol_S1En is: %10d\n"%(num_acc_optmols_INTERSECTIONs[-1]  -  num_acc_optmols_INTERSECTIONs[-2] ) )
            salam_log.write("The varation on num_acc_optmols_INTERSECTIONs is: %10d\n"%(num_acc_optmols_INTERSECTIONs[-1]  -  num_acc_optmols_INTERSECTIONs[-2]) )
        else:
            print('')
        
        salam_log.write("\n### End of run_accumulate_optmols_intersection.\n")  



#------------------------------------------------------------------------------
#------- main begin here ------------------------------------------------------
#------------------------------------------------------------------------------
if __name__ == '__main__':

    start_t0 = datetime.datetime.now()
    
    print("\n### Begin of program: SALAM.\n" )
    
    call_echo_begin_program_salam_to_salam_log = "echo -e  '### Begin of program: SALAM.\n'      1>    "    +   SALAM_LOG
    retcode = os.system( call_echo_begin_program_salam_to_salam_log )
    print("retcode = ", retcode)
    
    abundances = []
    num_replaceable_aromaticCHs = []
    num_acc_optmols = []
    num_acc_optmols_S1En = []
    num_acc_optmols_INTERSECTIONs = []
    
    
    #-------loop over PG002 ~ PG011 ---------------------------------------------------------
    print("### loop over PG002 ~ PG015. ")
    print("Initial MUTATION_GENERATION = %4d\n"% int( PARAS_DIC['MUTATION_GENERATION'] ) )
    print("MUTATION_GENERATION_LIMIT = %4d\n"% int(  PARAS_DIC['MUTATION_GENERATION_LIMIT'] ) )
    
    MUTATION_GENERATION = int( PARAS_DIC['MUTATION_GENERATION'] )
    
    while ( MUTATION_GENERATION  < int( PARAS_DIC['MUTATION_GENERATION_LIMIT'] )):
        print("### --------------------- MUTATION_GENERATION = %4d -------------------------\n"%MUTATION_GENERATION )
        retcode = os.system( "echo  '### --------------------- MUTATION_GENERATION =  %4d"%MUTATION_GENERATION  +  "   '      1>>    "    +   SALAM_LOG )
        print("retcode = ", retcode)
        
        GXXXX = 'G%04d'%MUTATION_GENERATION
        PXXXX = 'P%04d'%MUTATION_GENERATION
        OXXXX = 'O%04d'%MUTATION_GENERATION
        GXXXXPLUSONE = 'G%04d'%(MUTATION_GENERATION +1)
    
        # GXXXX_sdf_path = './project/'  +  GXXXX    +  '/'  + GXXXX   +  '.sdf'
        # all_frags = get_allfrags_via_bricsdecompose(filename=GXXXX_sdf_path)
        # # for frag in all_frags:
        # #     print( frag )
    
        GXXXX_sdf_path = './project/'  +  GXXXX    +  '/'  + GXXXX   +  '.sdf' 
        num_replaceable_aromaticCH = get_num_replaceable_aromatic_CHs(filename=GXXXX_sdf_path,
                                                                     patt_str='c[H]'
                                                                     )
        num_replaceable_aromaticCHs.append( num_replaceable_aromaticCH )
    
    
        if (MUTATION_GENERATION < RESTART_GENERATION):
            print("\nSkip all PGs, since MUTATION_GENERATION < RESTART_GENERATION.\n")
            
            # run_PG002()                                      #  './PG002__MaxMinPicker.py'
            # run_PG003()                                      #  './PG003__bat_sdf_to_com_gs.py'
            # run_PG004()                                      #  './PG004__bat_qsub_jobs_opt.py'
            
            #time.sleep( SLEEP_SECONDS )                      #  sleep before opt_analysis 
            # run_PG005(call_manner=1)                         #  './PG005__optimization_analysis.py'
            # run_PG006()                                      #  './PG006__bat_g09log_to_tadfcom.py'
            # run_PG007()                                      #  './PG007__bat_qsub_jobs_tadf.py'
            
            # time.sleep( SLEEP_SECONDS )                      #  sleep before prop_analysis 
            
            # run_PG008(call_manner=1)                        #  './PG008__property_analysis_completeness.py'
            # run_PG09(call_manner=2)                        #  './PG009__property_analysis_summary.py'
            # run_PG010(call_manner=2)                         #  './PG010__featurizers_train_evaluate.py'
            
            # run_PG011(call_manner=2)                         #  './PG011__predict_opt_molecules.py'
            # time.sleep( SLEEP_SECONDS )         
    
            # run_PG012(call_manner=2)                         #  './PG012__Generate_Gnew.py'
            # time.sleep( SLEEP_SECONDS )                      #  sleep before evaluate_abundance
            
            # run_PG013(call_manner=2)                         #   './PG013__MurckoScaffold.py'
            # time.sleep( SLEEP_SECONDS ) 
            
            # run_PG014(call_manner=2)                         #  './PG014__Energy_Sieve.py'
            # time.sleep( SLEEP_SECONDS ) 
            
            # run_PG015(call_manner=2)                         #  './PG015__Map-mols-by-FPSimilarity.py'
            # time.sleep( SLEEP_SECONDS )
                                  
        elif (MUTATION_GENERATION == RESTART_GENERATION):
            if (RESTART_MAXMINPICKER == True):
                run_PG002()                                      #  './PG002__MaxMinPicker.py'
            else:
                print("(MUTATION_GENERATION == RESTART_GENERATION), & (RESTART_MAXMINPICKER == False), skip run_PG002()")
            
            run_PG003()                                      #  './PG003__bat_sdf_to_com_gs.py'
            
            if (RESTART_OPTJOB == True):
                run_PG004()                                      #  './PG004__bat_qsub_jobs_opt.py'
            else:
                print("(MUTATION_GENERATION == RESTART_GENERATION), & (RESTART_OPTJOB == False), skip run_PG004()")
            
            time.sleep( SLEEP_SECONDS )                      #  sleep before opt_analysis 
            run_PG005(call_manner=1)                         #  './PG005__optimization_analysis.py'
    
            run_PG006()                                      #  './PG006__bat_g09log_to_tadfcom.py'
            
            if (RESTART_OPTJOB == True) or (RESTART_PROPJOB == True):
                run_PG007()                                      #  './PG007__bat_qsub_jobs_tadf.py'
            else:
                print("(MUTATION_GENERATION == RESTART_GENERATION), & ( (RESTART_OPTJOB == False) & (RESTART_PROPJOB == False) ), skip run_PG007()" )
                
            time.sleep( SLEEP_SECONDS )                      #  sleep before prop_analysis 
            
            run_PG008(call_manner=1)                        #  './PG008__property_analysis_completeness.py'
            run_PG009(call_manner=2)                        #  './PG009__property_analysis_summary.py'
            run_PG010(call_manner=2)                         #  './PG010__featurizers_train_evaluate.py'
            
            run_PG011(call_manner=2)                         #  './PG011__predict_opt_molecules.py'
            time.sleep( SLEEP_SECONDS )                      #  sleep before evaluate_abundance
            
            run_PG012(call_manner=2)                         #  './PG012__Generate_Gnew.py'
            time.sleep( SLEEP_SECONDS )
    
            run_PG013(call_manner=2)                         #   './PG013__MurckoScaffold.py'
            time.sleep( SLEEP_SECONDS )                    
            
            run_PG014(call_manner=2)                         #  './PG014__Energy_Sieve.py'
            time.sleep( SLEEP_SECONDS ) 
    
            run_PG015(call_manner=2)                         #  './PG015__Map-mols-by-FPSimilarity.py'
            time.sleep( SLEEP_SECONDS )
        
        else:
            # run_PG002()                                      #  './PG002__MaxMinPicker.py'
            # run_PG003()                                      #  './PG003__bat_sdf_to_com_gs.py'
            # run_PG004()                                      #  './PG004__bat_qsub_jobs_opt.py'
            
            # time.sleep( SLEEP_SECONDS )                      #  sleep before opt_analysis 
            # run_PG005(call_manner=1)                         #  './PG005__optimization_analysis.py'
            # run_PG006()                                      #  './PG006__bat_g09log_to_tadfcom.py'
            # run_PG007()                                      #  './PG007__bat_qsub_jobs_tadf.py'
            
            # time.sleep( SLEEP_SECONDS )                      #  sleep before prop_analysis 
            
            # run_PG008(call_manner=1)                        #  './PG008__property_analysis_completeness.py'
            # run_PG009(call_manner=2)                        #  './PG009__property_analysis_summary.py'
            # run_PG010(call_manner=2)                         #  './PG010__featurizers_train_evaluate.py'
            
            # run_PG011(call_manner=2)                         #  './PG011__predict_opt_molecules.py'
        
            # time.sleep( SLEEP_SECONDS )                      #  sleep before evaluate_abundance
            
            # run_PG012(call_manner=2)                         #  './PG012__Generate_Gnew.py'        
            # time.sleep( SLEEP_SECONDS )                   
            
            # run_PG013(call_manner=2)                         #   './PG013__MurckoScaffold.py'
            # time.sleep( SLEEP_SECONDS ) 
            
            # run_PG014(call_manner=2)                         #  './PG014__Energy_Sieve.py'
            # time.sleep( SLEEP_SECONDS ) 
            
            run_PG015(call_manner=2)                         #  './PG015__Map-mols-by-FPSimilarity.py'
            time.sleep( SLEEP_SECONDS )
        
        
        # run_evaluate_abundance(call_manner=2)            #  by call method: percentage_lt_criticval() in Metrics module
        # if (len(abundances) > 0) and ( abs( abundances[-1] - float(PARAS_DIC['ABUNANCE_LIMIT']) ) < 0.01):
        #     print("\nThe abundance has converged to ABUNANCE_LIMIT.\n")
        #     os.system("echo  '\nThe abundance has converged to ABUNANCE_LIMIT.\n'      1>>    "    +   SALAM_LOG)
        #     break
        # time.sleep( SLEEP_SECONDS )         
        
        # run_accumulate_optmols()
        # if ( (len(num_acc_optmols) > 0) and (num_acc_optmols[-1]  >= int(PARAS_DIC['NUM_ACC_OPTMOLS_LIMIT'])) ):
        #     print("\nThe num_acc_optmol has converged to NUM_ACC_OPTMOLS_LIMIT.\n")
        #     os.system("echo  '\nThe num_acc_optmol has converged to NUM_ACC_OPTMOLS_LIMIT.\n'      1>>    "    +   SALAM_LOG)
        #     break    
        # time.sleep( SLEEP_SECONDS )
        
        # run_accumulate_optmols_S1En( )
        # time.sleep( SLEEP_SECONDS )        
    
        MUTATION_GENERATION += 1
            
    
    
    #------------------------------------------------------------------------------
    MUTATION_GENERATION -= 1
    
    # #----- call sort_mols_by_SAScores ---------------------------------------------
    # run_sort_mols_by_SAScores()
    
    # #------ call accumulate_optmols_intersection -----------------------------------
    # run_accumulate_optmols_intersection( )
    
    
    #-----------------------------------------------------------------------------
    
    
    if len(num_replaceable_aromaticCHs) > 0:
        print("-"*60)
        print("\nnum_replaceable_aromaticCHs:")
        for i, num_replaceable_aromaticCH in zip( range(len(num_replaceable_aromaticCHs)),  num_replaceable_aromaticCHs):
            print("%8d %10.4f"%(i, num_replaceable_aromaticCH))
    else:
        print("\nlen(num_replaceable_aromaticCHs) > 0 is False.")
    
    
    if (len(abundances) > 0):
        print("-"*60)
        print("muation_generation, abundance are:")
        for i, abundance in zip( range(len(abundances)),  abundances):
            print("%8d %10.4f"%(i, abundance))
    else:
        print("\nlen(abundances) > 0 is False.")
    
    
    if (len(num_acc_optmols) > 0):
        print("-"*60)
        print("muation_generation, num_acc_optmols are:")
        for i, num_acc_optmol in zip( range(len(num_acc_optmols)),  num_acc_optmols):
            print("%8d %10d"%(i, num_acc_optmol))
    else:
        print("\nlen(num_acc_optmols) > 0 is False.")
    
    
    if (len(num_acc_optmols_S1En) > 0):
        print("-"*60)
        print("muation_generation, num_acc_optmols_S1En_R num_acc_optmols_S1En_G num_acc_optmols_S1En_B  are:")
        for i, num_acc_optmol_S1En in zip( range(len(num_acc_optmols_S1En)),  num_acc_optmols_S1En):
            print("%8d %8d %8d %8d"%(i, num_acc_optmol_S1En[0], num_acc_optmol_S1En[1], num_acc_optmol_S1En[2]))
    else:
        print("\nlen(num_acc_optmols_S1En) > 0 is False.")
    
    
    # #---- plot generation Vs abundance, aromaticCHs, num_acc_optmols pictures ----------------
    # df_generation_abundance_aromaticCHs_numaccoptmols = pd.DataFrame(data=zip( range(len(abundances)),  abundances, num_replaceable_aromaticCHs, num_acc_optmols), 
    #                                                                   columns=["generation", "abundance", "aromaticCH", "num_acc_optmols"]
    #                                                                   )
    
    # fig, ax = plt.subplots(1, 1, figsize=(12, 8), sharey=False)
    # df_generation_abundance_aromaticCHs_numaccoptmols.plot(
    #                                                         title='abundance vs generation',
    #                                                         #kind='scatter', 
    #                                                         kind='line', linestyle='-',
    #                                                         x='generation', 
    #                                                         y='abundance', 
    #                                                         color='g', 
    #                                                         alpha=0.95, 
    #                                                         ax=ax,
    #                                                         xlabel='generation', 
    #                                                         ylabel='abundance'
    #                                                         )
    
    # ax.legend(ncol=1, fontsize='x-large', shadow=True)
    # ax.grid(True)
    # fig.show()
    
    
    # fig1, ax1 = plt.subplots(1, 1, figsize=(12, 8), sharey=False)
    # df_generation_abundance_aromaticCHs_numaccoptmols.plot(
    #                                                         title='aromaticCH vs generation',
    #                                                         #kind='scatter', 
    #                                                         kind='line', linestyle='-',
    #                                                         x='generation', 
    #                                                         y='aromaticCH', 
    #                                                         color='r', 
    #                                                         alpha=0.95, 
    #                                                         ax=ax1, 
    #                                                         xlabel='generation',
    #                                                         ylabel='aromaticCH'
    #                                                         )
    
    # ax1.legend(ncol=1, fontsize='x-large', shadow=True)
    # ax1.grid(True)
    # fig1.show()
    
    # fig2, ax2 = plt.subplots(1, 1, figsize=(12, 8), sharey=False)
    # df_generation_abundance_aromaticCHs_numaccoptmols.plot(
    #                                                         title='num_acc_optmols vs generation',
    #                                                         #kind='scatter', 
    #                                                         kind='line', linestyle='-',
    #                                                         x='generation', 
    #                                                         y='num_acc_optmols', 
    #                                                         color='b', 
    #                                                         alpha=0.95, 
    #                                                         ax=ax2, 
    #                                                         xlabel='generation',
    #                                                         ylabel='num_acc_optmols'
    #                                                         )
    
    # ax2.legend(ncol=1, fontsize='x-large', shadow=True)
    # ax2.grid(True)
    # fig2.show()
    
    
    #----- plot_muation_diagram --------------------------------------------------
    plot_muation_diagram(filename='./project/mutation_diagram.csv',
                         figname='./project/muation-diagram.png',
                         molsPerRow=3,
                         subImgSize=(800, 800),
                         num_of_mols=None
                         )
    
    
    
    print("\n### End of program: SALAM.\n" )
    
    call_echo_end_program_salam_to_salam_log = "echo -e '\n### END of program: SALAM.\n'      1>>    "    +   SALAM_LOG
    retcode = os.system( call_echo_end_program_salam_to_salam_log )
    print("retcode = ", retcode)
    
    
    end_t0 = datetime.datetime.now()
    elapsed_sec0 = (end_t0 - start_t0).total_seconds()
    print("\nTotal running time of SALAM is : " + "{:.2f}".format(elapsed_sec0/60.) + " min ")
    
    retcode = os.system( "echo  '\nTotal running time of SALAM is :  " +  "{:.2f}".format(elapsed_sec0/60.) + " min   '  "  +  "     1>>    "    +   SALAM_LOG )
    print("retcode = ", retcode)
    
    print("\n### Normal termination of program: SALAM!\n")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Apr 16 10:15:03 2021

@author: tucy
"""

import re
import glob
#import numpy as np
#import pandas as pd

#from rdkit import Chem
#from rdkit.Chem import AllChem
import subprocess
#import shutil
import os
import argparse
import time
#import subprocess
#from sys import exit as sys_exit

parser = argparse.ArgumentParser(description="manual to this script:")
#parser.add_argument('--analyzed_property', type=str, default="tadf")
parser.add_argument('--num_serviers', type=int, default=2)
parser.add_argument('--remote_wkdir', type=str, default="project")
parser.add_argument('--pbs_job_incompleteness_tolerance', type=float, default=0.01)
parser.add_argument('--sleep_seconds', type=int, default=300)
parser.add_argument('--GXXXX', type=str, default="G0000")
parser.add_argument('--PXXXX', type=str, default="P0000")
parser.add_argument('--OXXXX', type=str, default="O0000")

args = parser.parse_args()

#print("opt_property: ", args.analyzed_property, type(args.analyzed_property) )

print("pbs_job_incompleteness_tolerance: ", args.pbs_job_incompleteness_tolerance, type(args.pbs_job_incompleteness_tolerance) )

#analyzed_property = args.analyzed_property
pbs_job_incompleteness_tolerance = args.pbs_job_incompleteness_tolerance

print(args.GXXXX, type(args.GXXXX))
print(args.PXXXX, type(args.PXXXX))
print(args.OXXXX, type(args.OXXXX))
GXXXX = args.GXXXX
PXXXX = args.PXXXX
OXXXX = args.OXXXX

print(args.sleep_seconds, type(args.sleep_seconds))
SLEEP_SECONDS = args.sleep_seconds

print("num_serviers: ", args.num_serviers, type(args.num_serviers))
NUM_SERVIERS = args.num_serviers

print(args.remote_wkdir, type(args.remote_wkdir))
remote_wkdir = args.remote_wkdir


#-----------------------------------------------------------------------------
def compute_num_total_finished_jobs():
    
    com_filenames = glob.glob(r'./cpd*.com')
    filenames = glob.glob(r'./cpd*.log')
    
    if (len(com_filenames) == 0):
        print("No cpd*.com file exit! SYSTEM EXIT(101) in compute_num_total_finished_jobs() of PG005.")
        exit(101)
    else:
        if (len(filenames) == 0):
            print("No cpd*.log file exit! The ruturn is set to 0.")
            return len(com_filenames), 0
        else:
            failed_compound_idxs = []
            
            idx = 0
            for filename  in  filenames:
                termination_pattern = re.compile(r'Normal(\s*)termination(.*?)', re.MULTILINE|re.DOTALL)
                with open(filename) as fp:
                    fc=fp.read()
                    termination_result = termination_pattern.findall(fc)
                    #print("# ", filename, "", len(termination_result))
                    if len(termination_result) != 1:
                        failed_compound_idxs.append(idx)
                    
                    idx += 1


            com_compoundnames = [x.replace(r'./cpd-', '').replace(r'.com', '')   for  x in com_filenames]
            compoundnames = [x.replace(r'./cpd-', '').replace(r'.log', '')   for  x in filenames]            
            
            if len(failed_compound_idxs) > 0:
                print('-'*60)
                print("failed_compound_names: ")
                for failed_compound_idx in failed_compound_idxs:            
                    print(filenames[failed_compound_idx])
                
                failed_compound_idxs.sort(reverse=True)
                for i  in failed_compound_idxs:
                    compoundnames.pop(i)
            
            # print('-'*60)
            # for name in compoundnames:
            #     print(name)
            
                print("len(com_compoundnames), len(compoundnames) = ", len(com_compoundnames), len(compoundnames) )
                return len(com_compoundnames), len(compoundnames)
            else:
                print("len(com_compoundnames), len(compoundnames) = ", len(com_compoundnames), len(compoundnames) )
                return len(com_compoundnames), len(compoundnames)
                


def compute_pbs_job_completeness(pbs_jobs_wkdir='./project/%s/pick_mols/com-files/finished/com-files/'%GXXXX, 
                                 project_dir='../../../../../../'):
    os.chdir( pbs_jobs_wkdir ) 
    #CWD = os.getcwd()
    #print("CWD: \n", CWD)
    
    num_total_jobs, num_finished_jobs = compute_num_total_finished_jobs()
    
    job_completeness = 0.
    
    if (num_total_jobs != 0):
        job_completeness = num_finished_jobs / float( num_total_jobs )
    else:
        print("No cpd-*.com files exist!!!")
        exit(100)
    
    os.chdir( project_dir ) 
    #CWD = os.getcwd()
    #print("CWD: \n", CWD)
    
    return job_completeness   
    


def get_normal_lognames(pbs_jobs_wkdir='./project/%s/pick_mols/com-files/finished/com-files/'%GXXXX, 
                        project_dir='../../../../../../'):
    
    os.chdir( pbs_jobs_wkdir ) 
    
    filenames = glob.glob(r'./cpd*.log')
    
    if (len(filenames) == 0):
        print("No cpd*.log file exit! The ruturn is set to 0.")
        exit(1)
        return 
    else:
        failed_log_idxs = []
        
        idx = 0
        for filename  in  filenames:
            termination_pattern = re.compile(r'Normal(\s*)termination(.*?)', re.MULTILINE|re.DOTALL)
            with open(filename) as fp:
                fc=fp.read()
                termination_result = termination_pattern.findall(fc)
                #print("# ", filename, "", len(termination_result))
                if len(termination_result) != 1:
                    failed_log_idxs.append(idx)
                
                idx += 1
        
        normal_logfile_names = [x.replace(r'./', '').replace(r'.log', '')   for  x in filenames]
        
        if len(failed_log_idxs) > 0:
            print('-'*60)
            print("# failed_compound_names: ")
            for failed_log_idx in failed_log_idxs:            
                print(filenames[failed_log_idx])
        
            
            failed_log_idxs.sort(reverse=True)
            print("# pop failed log names:")
            for i  in failed_log_idxs:
                normal_logfile_names.pop(i)
                print("idx = ", i)
        
        
        print("len(normal_logfile_names) = ", len(normal_logfile_names))
        
        os.chdir( project_dir ) 
    
        return normal_logfile_names



def synchronize_job_dir(GXXXX, 
                        analysis_dir = './project/'  +  GXXXX  +  '/pick_mols/com-files/',
                        GXXXX_opt_jobs_dir =  './project/'  +  GXXXX  +  '/pick_mols/com-files/' + 'opt_jobs/', 
                        remote_GXXXX_opt_jobs_dir = '/home/tucy/Computation-node131/Gaussian16-jobs/SALAM/project/'  + GXXXX +  "/opt_jobs/", 
                        base_wkdir = '../../../../',
                        NUM_SERVIERS = NUM_SERVIERS
                        ):

    print("Begin synchronize_job_dir")
    os.chdir(analysis_dir)
    
    if (NUM_SERVIERS == 2):
        print("NUM_SERVIERS = ", NUM_SERVIERS)
        status, output = subprocess.getstatusoutput(' mkdir   ./remote_prop_jobs/   2>/dev/null  ' )
        status, output = subprocess.getstatusoutput(' scp  -r  tucy@node130:' + remote_GXXXX_opt_jobs_dir  + 'cpd*.log '  +  '    ./remote_prop_jobs/ ' )
        print("scp -r remote_GXXXX_opt_jobs_dir to  ./remote_prop_jobs/. status, output = ", status, output)
    else:
        print("NUM_SERVIERS = ", NUM_SERVIERS)
        
    
    logfilenames = glob.glob(r'./*prop_jobs/cpd*.log')
    
    if (len(logfilenames) == 0):
        print("No cpd*.log file exit!")
    else:
        for logfilename  in  logfilenames:
            termination_pattern = re.compile(r'Normal(\s*)termination(.*?)', re.MULTILINE|re.DOTALL)
            with open(logfilename) as fp:
                fc=fp.read()
                termination_result = termination_pattern.findall(fc)
                #print("# ", filename, "", len(termination_result))
                if len(termination_result) == 1:
                    status, output = subprocess.getstatusoutput(' cp   ' + logfilename +   '  ./    2>/dev/null  ' )
                    print("cp logfile. status, output = ", status, output)

    
    os.chdir( base_wkdir )
    time.sleep(120)
    
    print("End synchronize_job_dir")
    
    return 0 




#-----------------------------------------------------------------------------

#--------main func -----------------------------------------------------------
#-----------------------------------------------------------------------------

base_wkdir = os.getcwd()
print("base_wkdir: \n", base_wkdir)


#-------main loop ------------------------------------------------------------
print('*'*60)
print("# PG008: Begin compute pbs_job_completeness, only when attaining the desired job_completeness, the loop could be stopped and turnned to next step.")


synchronize_job_dir(GXXXX, 
                    analysis_dir = './project/'  +  GXXXX  +  '/pick_mols/com-files/finished/com-files/',
                    GXXXX_opt_jobs_dir =  './project/'  +  GXXXX  +  '/pick_mols/com-files/finished/com-files/' + 'prop_jobs/', 
                    remote_GXXXX_opt_jobs_dir = '/home/tucy/Computation-node131/Gaussian16-jobs/SALAM/%s/'%remote_wkdir  + GXXXX +  "/prop_jobs/", 
                    base_wkdir = base_wkdir,
                    NUM_SERVIERS = NUM_SERVIERS
                    )


GXXXX_pick_mols_com_files_finished_com_files_dir =  './project/'  +   GXXXX  +  '/pick_mols/com-files/finished/com-files/'
job_completeness = compute_pbs_job_completeness(pbs_jobs_wkdir=GXXXX_pick_mols_com_files_finished_com_files_dir,  
                                                project_dir=base_wkdir)

print("The initial job_completeness = ", job_completeness)

job_completeness_summary_file_path = './project/'   +   GXXXX   +   '/pick_mols/com-files/finished/com-files/job_completeness_summary.data'

loop_times = 1
with open(job_completeness_summary_file_path, 'w') as tmp:
    while ( abs( 1. - job_completeness ) >  pbs_job_incompleteness_tolerance ):
        print("The job_completeness is %8.4f." % job_completeness)
        print("loop_times = %d. "%(loop_times) )
        #print("Sleep 1 hour.")
        #time.sleep(3600)
        
        synchronize_job_dir(GXXXX, 
                            analysis_dir = './project/'  +  GXXXX  +  '/pick_mols/com-files/finished/com-files/',
                            GXXXX_opt_jobs_dir =  './project/'  +  GXXXX  +  '/pick_mols/com-files/finished/com-files/' + 'prop_jobs/', 
                            remote_GXXXX_opt_jobs_dir = '/home/tucy/Computation-node131/Gaussian16-jobs/SALAM/%s/'%remote_wkdir  + GXXXX +  "/prop_jobs/", 
                            base_wkdir = base_wkdir,
                            NUM_SERVIERS = NUM_SERVIERS
                            )
        
        print("Sleep %d sec."%SLEEP_SECONDS )
        time.sleep(SLEEP_SECONDS)
        job_completeness = compute_pbs_job_completeness(pbs_jobs_wkdir=GXXXX_pick_mols_com_files_finished_com_files_dir, 
                                                        project_dir=base_wkdir)
        
        tmp.write( "%10d, "%(loop_times) )
        tmp.write( "%15.6f\n"%(job_completeness) )
        loop_times += 1

with open(job_completeness_summary_file_path, 'a+') as tmp:
    tmp.write( "%15.6f\n"%(job_completeness) )
    
    
job_completeness  = compute_pbs_job_completeness(pbs_jobs_wkdir=GXXXX_pick_mols_com_files_finished_com_files_dir, 
                                                 project_dir=base_wkdir
                                                )
    
print("The final job_completeness is %8.4f.\n" % job_completeness)

#-----------------------------------------------------------------------------

os.chdir(base_wkdir)
CWD = os.getcwd()
print("\nCWD: \n", CWD)
#------------------------------------------------------------------------------



print("\n### End of program:  PG008!\n")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov 10 16:05:54 2021

@author: tucy
"""

import subprocess
import os

import argparse

parser = argparse.ArgumentParser(description="manual to this script:")
parser.add_argument('--num_serviers', type=int, default=2)
parser.add_argument('--remote_wkdir', type=str, default="project")
parser.add_argument('--GXXXX', type=str, default="G0000")
parser.add_argument('--PXXXX', type=str, default="P0000")
parser.add_argument('--OXXXX', type=str, default="O0000")

parser.add_argument('--COMPUTE_RATIO_SERVIER1', type=float, default=0.56)
parser.add_argument('--COMPUTE_RATIO_SERVIER2', type=float, default=0.46)

args = parser.parse_args()


print("num_serviers: ", args.num_serviers, type(args.num_serviers))
NUM_SERVIERS = args.num_serviers

print(args.remote_wkdir, type(args.remote_wkdir))
remote_wkdir = args.remote_wkdir

print(args.GXXXX, type(args.GXXXX))
print(args.PXXXX, type(args.PXXXX))
print(args.OXXXX, type(args.OXXXX))
GXXXX = args.GXXXX
PXXXX = args.PXXXX
OXXXX = args.OXXXX

G_num = int( GXXXX[1:] )  
print("G_num = ", G_num)

print(args.COMPUTE_RATIO_SERVIER1, type(args.COMPUTE_RATIO_SERVIER1))
print(args.COMPUTE_RATIO_SERVIER2, type(args.COMPUTE_RATIO_SERVIER2))
COMPUTE_RATIO_SERVIER1 = args.COMPUTE_RATIO_SERVIER1
COMPUTE_RATIO_SERVIER2 = args.COMPUTE_RATIO_SERVIER2



#-------- remote servier working dir ------------------------------------------
remote_base_wkdir = '/home/tucy/Computation-node131/Gaussian16-jobs/SALAM/%s/'%remote_wkdir
remote_GXXXX_prop_jobs_dir = remote_base_wkdir  + GXXXX +  "/prop_jobs/" 

print("The remote_GXXXX_prop_jobs_dir:")
print(remote_GXXXX_prop_jobs_dir)


#------------------------------------------------------------------------------
base_wkdir = os.getcwd()
print("base_wkdir: \n", base_wkdir)

GXXXX_pick_mols_com_files_finished_com_files_dir = './project/' +  GXXXX  + '/pick_mols/com-files/finished/com-files/prop_jobs/'
os.chdir( GXXXX_pick_mols_com_files_finished_com_files_dir )
CWD = os.getcwd()
print("CWD: \n", CWD)


if (NUM_SERVIERS == 2):
    print("NUM_SERVIERS = ", NUM_SERVIERS)
    retcode = subprocess.call('./bat-g16-prop.sh  '  +  str(COMPUTE_RATIO_SERVIER1)    +  '   %d  '%G_num , shell=True)
    #retcode = subprocess.call('./bat-g16-prop.sh ', shell=True)
    print(retcode)
    
    #------ run remote pbs jobs ---------------------------------------------------
    status, output = subprocess.getstatusoutput('ssh    tucy@node130    "cd   '   +   remote_GXXXX_prop_jobs_dir   +  ' ;    ./bat-g16-prop_node130_133.sh     '  +  str(COMPUTE_RATIO_SERVIER2)    +  '   %d  '%G_num   +  '     > /dev/null   2>&1   &  "  ')  
    print("run remote pbs_jobs status. status, output = ", status, output)

else:
    print("NUM_SERVIERS = ", NUM_SERVIERS)
    retcode = subprocess.call('./bat-g16-prop.sh   1 '   +  '   %d  '%G_num  , shell=True)


#status, output = subprocess.getstatusoutput('qstat -u tucy ')
status, output = subprocess.getstatusoutput('  qstat -u tucy  | grep "\<0 R\>"  -C1   ;   echo   "" ;    qstat -u tucy | grep -i "tucy"  -c ;    ')
print("pbs_jobs status. status, output = ", status, output)


os.chdir(base_wkdir)
CWD = os.getcwd()
print("CWD: \n", CWD)
#------------------------------------------------------------------------------


print("\n### End of program: PG007!\n")
